/*
	$Id $

	Copyright (C) 2016 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.hibernate.mapping.PersistentClass;

import net.sf.basedb.core.data.AnnotatableData;
import net.sf.basedb.core.data.LoggableData;
import net.sf.basedb.core.log.ChangeType;
import net.sf.basedb.core.log.LoggingInterceptor;
import net.sf.basedb.core.log.db.AnnotationLogger;
import net.sf.basedb.core.snapshot.SnapshotManager;
import net.sf.basedb.util.EqualsHelper;
import net.sf.basedb.util.units.UnitConverter;

/**
	Batcher implementation for adding, updating and removing multiple 
	annotations to a large number of items in a single transaction
	while keeping the memory usage as low as possible.
	The batcher bypasses Hibernate where it is possible and uses direct
	SQL to access the database. There are some limitations:
	
	<ul>
	<li>It can only work with items that are already in the database. It is not
		possible to use the batcher with items that are created in the same 
		transaction.

	<li>It can only work with items of a single {@link Item} type.

	<li>A item can only be processed once (eg. {@link #setCurrentItem(Annotatable)} can
		only be called once for a given item.
	
	<li>{@link Annotatable#getAnnotationSet()} must not have been called on the item
		before using it with this batcher.
		
	<li>The annotation types to work with must be specified before the
		first call to {@link #setCurrentItem(Annotatable)}. It is not possible 
		to work with other annotations.
		
	<li>The batcher only works with {@link Annotation.Source#PRIMARY} annotations.
	
	<li>The {@link #setValues(AnnotationType, List, Unit, boolean)} can only be called once
		per item and annotation type.
	
	</ul>
	
	@author nicklas
	@since 3.8
*/
public class AnnotationBatcher 
	extends AbstractBatcher
{
	
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger("net.sf.basedb.core.AnnotationBatcher");
	private final boolean isDebugEnabled = log.isDebugEnabled();
	
	private final Item itemType;
	
	// All updates through this batcher get the same date+time
	private final Date batchDate;

	// We need some interaction with Hibernate
	private final org.hibernate.StatelessSession statelessSession;
	private final org.hibernate.dialect.Dialect dialect;
	private final org.hibernate.Cache secondLevelCache;

	// Log changes directly to the LoggingInterceptor
	private final LoggingInterceptor loggingInterceptor;

	// Discriminator values for Type items (eg. Type.INT -> IntegerParameterValueData -> 1)
	// Pre-loaded and used when inserting into "ParameterValues" table
	private final int[] discriminators;
		
	// SQL query through Hibernate to load information about current annotations
	// for a single item. Uses 'annotationSetId' as the only input parameter.
	private final org.hibernate.Query loadAnnotationInfo;
	
	// SQL query through Hibernate to load annotation values
	// One query for each type of value (eg. Type.INT, Type.STRING, etc.)
	// Uses a list of ParameterValue.id as input parameter so we can load
	// values for multiple annotations with a single query
	private final org.hibernate.Query[] loadAnnotationValues;

	// Batcher for inserting rows into the "ParameterValues" table
	private final InsertBatcher insertIntoParameterValues;
	// Batcher for inserting rows into the "Annotations" table
	private final InsertBatcher insertIntoAnnotations;
	// Batcher for updating rows in the "Annotations" table
	private final UpdateBatcher updateAnnotations;
	// Batcher for inserting rows into the "AnnotationSets" table
	private final InsertBatcher insertIntoAnnotationSets;
	// Batcher for updating the item table when a new annotation set has been created
	private final UpdateBatcher updateItemTable;

	// All batchers (to make it easy to flush/close all at once)
	private final List<UpdateBatcher> allUpdateBatchers;
	private final List<InsertBatcher> allInsertBatchers;
	
	private final UpdateBatcher deleteFromAnnotations;
	private final UpdateBatcher deleteFromParameterValues;
	private final UpdateBatcher nullifyClonedReferences;
	private final UpdateBatcher deleteInheritedAnnotations;
	private final UpdateBatcher[] deleteValues;
	private final UpdateBatcher[] insertValues;
	
	// The currently active item
	private int numItems;
	private Annotatable currentItem;
	private int currentAnnotationSetId;
	private boolean deleteSnapshotIfModifed;
	private boolean evictFromSecondLevelCacheIfModified;
	
	// Lists for temporary keeping track of value IDs for
	// each Type so we can use a single query for all values of
	// the same Type
	private final List<Integer>[] valueIds;
	
	// Maps AnnotationType.id -> Annotation information for current item
	private Map<Integer, CurrentAnnotationInfo> currentInfo;
	
	// List of AnnotationSet IDs that should be deleted
	// from the snapshot manager
	private final List<Integer> snapshotsToDelete;
	
	// Cache with unit converters so that we don't need to create
	// new ones all the time
	private final Map<String, UnitConverter> unitConvertes;
	
	/**
		Create a new batcher instance. 
		@param dc A DbControl to use for database access
		@param itemType The type of items to work with. This must be an annotatable item type.
	*/
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public AnnotationBatcher(DbControl dc, Item itemType)
	{
		if (dc == null) throw new NullPointerException("dc");
		if (itemType == null) throw new NullPointerException("itemType");
		if (!Metadata.getAnnotatableItems().contains(itemType))
		{
			throw new InvalidDataException(itemType+" is not an annotatable item class.");
		}
		
		this.itemType = itemType;
		this.batchDate = new Date();
		
		// Get the table the items are stored in since we may have to set 
		// 'annotationset_id' column for items that doesn't have any annotations
		String itemTypeTableName = HibernateUtil.getClassMapping(itemType.getDataClass().getName()).getTable().getName();

		this.statelessSession = dc.getStatelessSession();
		this.dialect = HibernateUtil.getDialect();
		this.secondLevelCache = HibernateUtil.getSessionFactory().getCache();
		// Log changes if the items are implementing the LoggableData interface
		this.loggingInterceptor = LoggableData.class.isAssignableFrom(itemType.getDataClass()) ? 
			dc.getLoggingInterceptor() : null;

		Type[] types = Type.values();
		// Load discriminator values from Hibernate metadata
		this.discriminators = new int[types.length];
		for (Type t : types)
		{
			PersistentClass pc = HibernateUtil.getClassMapping(t.newParameterValueData().getClass().getName());
			this.discriminators[t.ordinal()] = Integer.parseInt(pc.getDiscriminatorValue());
		}
		
		// For storing information about the currently selected item
		// Maps AnnotationType.id -> Annotation information
		this.currentInfo = new HashMap<Integer, CurrentAnnotationInfo>();
		this.unitConvertes = new HashMap<String, UnitConverter>();
		
		this.allUpdateBatchers = new ArrayList<UpdateBatcher>();
		this.allInsertBatchers = new ArrayList<InsertBatcher>();
		this.snapshotsToDelete = new ArrayList<Integer>();
		try
		{
			Connection c = HibernateUtil.getConnection(dc.getHibernateSession());
			
			// Initialize statements for loading information about current annotations
			this.loadAnnotationInfo = HibernateUtil.createSqlQuery(statelessSession, 
				PredefinedQuery.getQueryString("AB_LOAD_ANNOTATION_INFO"));
			this.loadAnnotationValues = new org.hibernate.Query[types.length];
			this.valueIds = new List[types.length];
			for (Type t : types)
			{
				this.valueIds[t.ordinal()] = new ArrayList<Integer>();
				this.loadAnnotationValues[t.ordinal()] = HibernateUtil.createSqlQuery(statelessSession, 
					PredefinedQuery.getQueryString("AB_LOAD_ANNOTATION_VALUES", t.getTableName()));
			}
			
			// Initialize statements for creating new annotations
			this.insertIntoParameterValues = createInsertBatcher(c,
				PredefinedQuery.getQueryString("AB_INSERT_INTO_PARAMETERVALUES"), 
				Types.INTEGER /* discriminator */
				);
			
			this.insertIntoAnnotations = createInsertBatcher(c,
				PredefinedQuery.getQueryString("AB_INSERT_INTO_ANNOTATIONS"),
				Types.INTEGER, /* annotationset_id */
				Types.INTEGER, /* annotationtype_id */
				Types.INTEGER, /* unit_id */
				Types.INTEGER, /* value_id */
				Types.TIMESTAMP /* last_updated */
				);

			this.updateAnnotations = createUpdateBatcher(c,
				PredefinedQuery.getQueryString("AB_UPDATE_ANNOTATIONS"),
				Types.INTEGER, /* version (new) */
				Types.INTEGER, /* unit_id */
				Types.TIMESTAMP, /* last_updated */
				Types.INTEGER, /* id */
				Types.INTEGER /* version (old) */
				);
			
			this.insertIntoAnnotationSets = createInsertBatcher(c,
				PredefinedQuery.getQueryString("AB_INSERT_INTO_ANNOTATIONSETS", Integer.toString(itemType.getValue())),
				Types.INTEGER /* item_id */
				);
			
			this.updateItemTable = createUpdateBatcher(c, 
				PredefinedQuery.getQueryString("AB_UPDATE_ITEMTABLE", itemTypeTableName),
				Types.INTEGER, /* annotationset_id */
				Types.INTEGER /* item.id */
				);
			
			// Initialize statements for deleting existing annotation
			// NOTE! The order here is important since the batchers
			// will be flused in the order they are defined and we want
			// to avoid foreign key errors
			this.nullifyClonedReferences = createUpdateBatcher(c, 
				PredefinedQuery.getQueryString("AB_NULLIFY_CLONED_ANNOTATIONS"),
				Types.INTEGER /* inherited_id */
				);
			
			this.deleteInheritedAnnotations = createUpdateBatcher(c, 
				PredefinedQuery.getQueryString("AB_DELETE_INHERITED_ANNOTATIONS"),
				Types.INTEGER /* inherited_id */
				);
			
			this.deleteFromAnnotations = createUpdateBatcher(c, 
				PredefinedQuery.getQueryString("AB_DELETE_FROM_ANNOTATIONS"), 
				Types.INTEGER /* annotation.id */
				);

			this.deleteValues = new UpdateBatcher[types.length];
			this.insertValues = new UpdateBatcher[types.length];
			for (Type t : types)
			{
				String valuesTableName = t.getTableName();

				// IMPORTANT! The DELETE must be before the INSERT or the 
				// batecher will insert new values before deleting old (and also delete the new)
				this.deleteValues[t.ordinal()] = createUpdateBatcher(c, 
					PredefinedQuery.getQueryString("AB_DELETE_FROM_VALUES", valuesTableName), 
					Types.INTEGER /* ParameterValue.id */
					);
				
				this.insertValues[t.ordinal()] = createUpdateBatcher(c, 
					PredefinedQuery.getQueryString("AB_INSERT_INTO_VALUES", valuesTableName), 
					Types.INTEGER, /* ParameterValue.id */ 
					t.getSQLType() /* value */
					);
			}

			this.deleteFromParameterValues = createUpdateBatcher(c, 
				PredefinedQuery.getQueryString("AB_DELETE_FROM_PARAMETERVALUES"), 
				Types.INTEGER /* id */
				);
		}
		catch (SQLException ex)
		{
			throw new BaseException(ex);
		}
		
		setDbControl(dc);
	}
	
	
	/*
	 	From the Batcher interface
	 	--------------------------
	*/
	
	/**
		Flush the batchers.
	*/
	@Override
	public void flush()
	{
		log.debug(numItems + " done. Flushing batchers.");
		String currentSql = null;
		try
		{
			for (UpdateBatcher b : allUpdateBatchers)
			{
				currentSql = b.getSql();
				b.flush();
			}
		}
		catch (SQLException ex)
		{
			log.error(currentSql, ex);
			LogUtil.logSQLExceptionChain(log, ex);
			throw new BaseException(ex);
		}
	}
	// ------------------------
	
	/*
		From the AbstractBatcher class
		------------------------------
	*/
	@Override
	protected void onBeforeClose()
	{
		log.debug(numItems + " done. Closing.");
		for (UpdateBatcher b : allUpdateBatchers)
		{
			b.close();
		}
		allUpdateBatchers.clear();
		for (InsertBatcher b : allInsertBatchers)
		{
			b.close();
		}
		allInsertBatchers.clear();
		if (isDebugEnabled) 
		{
			log.debug("Deleting annotation snapshots: " + snapshotsToDelete);
		}
		for (Integer id : snapshotsToDelete)
		{
			SnapshotManager.removeSnapshot(id);
		}
		snapshotsToDelete.clear();
		unitConvertes.clear();
		currentInfo.clear();
		currentItem = null;
	}
	// ---------------------------
	
	private InsertBatcher createInsertBatcher(Connection c, String sql, int... types)
		throws SQLException
	{
		sql = sql.replace('[', dialect.openQuote()).replace(']', dialect.closeQuote());
		InsertBatcher b = new InsertBatcher(c, sql, types);
		allInsertBatchers.add(b);
		return b;
	}
	
	private UpdateBatcher createUpdateBatcher(Connection c, String sql, int... types)
		throws SQLException
	{
		sql = sql.replace('[', dialect.openQuote()).replace(']', dialect.closeQuote());
		UpdateBatcher b = new UpdateBatcher(c, sql, types);
		allUpdateBatchers.add(b);
		return b;
	}
	
	/**
		Get/create a unit converter for converting between two units.
	*/
	private UnitConverter getUnitConverter(Unit from, Unit to)
	{
		String key = from.getId() + ":" + to.getId();
		UnitConverter converter = unitConvertes.get(key);
		if (converter == null)
		{
			converter = to.getUnitConverter(from);
			unitConvertes.put(key, converter);
		}
		return converter;
	}
	
	/**
		Add annotation types to work with. It is possible to call this method multiple
		times, but it must be done before {@link #setCurrentItem(Annotatable)} is called.
		The annotation types must be enabled for use with the item type this batcher
		is working with and the current user must have USE permission for all annotation 
		types.
	*/
	public void addAnnotationTypes(Collection<AnnotationType> types)
	{
		if (isClosed()) throw new ConnectionClosedException();
		if (currentItem != null)
		{
			throw new IllegalStateException("Not allowed after calling setCurrentItem().");
		}
		if (types == null || types.isEmpty()) return;
	
		for (AnnotationType at : types)
		{
			if (currentInfo.containsKey(at.getId())) continue; // Already added
			
			log.debug("Adding annotation type: " + at);
			
			at.checkPermission(Permission.USE);
			if (!at.isEnabledForItem(itemType))
			{
				throw new InvalidAnnotationException(at, itemType);
			}
			
			CurrentAnnotationInfo info = new CurrentAnnotationInfo(at, itemType);
			currentInfo.put(at.getId(), info);
		}
	}
	
	/**
		Set the current item to work with from now on. Before calling this method 
		the {@link #addAnnotationTypes(Collection)} must have specified the annotation
		types to work with.
		
		The item must not have been used with this method before and 
		the {@link Annotatable#getAnnotationSet()} method must not have been 
		called on the item. 
		
		If the item can be used, the batcher will immediately load all existing
		annotations and their values and prepare for adding, updating or removing
		values.
	*/
	public void setCurrentItem(Annotatable item)
	{
		if (isClosed()) throw new ConnectionClosedException();
		if (currentInfo.size() == 0)
		{
			throw new IllegalStateException("Must call addAnnotationTypes() before calling this method.");
		}
		
		if (item == null) throw new NullPointerException("item");
		if (item.getType() != itemType) throw new InvalidDataException("Item is not a " + itemType + ": " + item);
		if (item.getId() == 0) throw new InvalidDataException("Item has not been saved to database: " + item);
		((BasicItem)item).checkBatchAnnotatableUsage(false, true);
		item.checkPermission(Permission.WRITE);
		
		log.debug("Setting current item: " + item);
		
		if (numItems > 0 && numItems % getBatchSize() == 0)
		{
			flush();
		}
		
		// Reset old information and prepare for next item
		this.numItems++;
		this.currentItem = item;
		this.currentAnnotationSetId = 0;
		this.deleteSnapshotIfModifed = false;
		this.evictFromSecondLevelCacheIfModified = secondLevelCache != null && secondLevelCache.containsEntity(itemType.getDataClass(), item.getId());

		int itemId = item.getId();
		for (CurrentAnnotationInfo info : currentInfo.values())
		{
			info.itemId = itemId;
			info.hasBeenUsed = false;
			info.annotationId = null;
			info.version = 0;
			info.unitId = null;
			info.valueId = 0;
			info.lastUpdated = null;
			info.values.clear();
			info.newValues = null;
		}
		
		// Read current values
		if (item.isAnnotated())
		{
			currentAnnotationSetId = ((AnnotatableData)((BasicItem)item).getData()).getAnnotationSet().getId();
			loadAnnotationInfo.setInteger("annotationSet", currentAnnotationSetId);
			deleteSnapshotIfModifed = true;
			
			if (isDebugEnabled)
			{
				log.debug("Loading annotations for current item: " + item);
			}
			
			// Maps value_id -> Annotation info
			Map<Integer, CurrentAnnotationInfo> tmp = new HashMap<Integer, CurrentAnnotationInfo>();
			List<Object[]> list = HibernateUtil.loadList(Object[].class, loadAnnotationInfo, null);

			if (isDebugEnabled)
			{
				log.debug("Found " + list.size() + " annotations for current item: " + item);
			}
			
			for (Object[] data : list)
			{
				Integer annotationTypeId = (Integer)data[0];
				CurrentAnnotationInfo info = currentInfo.get(annotationTypeId);
				if (info == null) 
				{
					// This annotation type is not among the specified annotation types
					continue; // with the next annotation
				}
				
				if (isDebugEnabled)
				{
					log.trace("Found annotation: AnnotationType=" + info.annotationType);
				}
				
				// Index into arrays that depend on Type
				int typeIndex = info.valueType.ordinal();
				
				// Set current annotation information
				info.annotationId = (Integer)data[1];
				info.version = (Integer)data[2];
				info.unitId = (Integer)data[3];
				info.valueId = (Integer)data[4];
				info.lastUpdated = (Date)data[5];
				
				valueIds[typeIndex].add(info.valueId);
				tmp.put(info.valueId, info);
			}
			
			// Load values one Type at a time
			for (Type t : Type.values())
			{
				int typeIndex = t.ordinal();
				if (valueIds[typeIndex].isEmpty()) continue; // No annotation with this value type
				
				if (isDebugEnabled)
				{
					log.debug("Loading " + t.name() + " values: " + valueIds[typeIndex]);
				}
				
				org.hibernate.Query query = loadAnnotationValues[typeIndex];
				query.setParameterList("listOfIds", valueIds[typeIndex], Type.INT.getTypeWrapper().getHibernateType());
			
				list = HibernateUtil.loadList(Object[].class, query, null);
				if (isDebugEnabled)
				{
					log.debug("Found " + list.size() + " values");
				}
				for (Object[] data : list)
				{
					Integer valueId = (Integer)data[0];
					tmp.get(valueId).values.add(data[1]);
					if (isDebugEnabled) log.trace(valueId + ": "+data[1]);
				}
				
				valueIds[typeIndex].clear();
			}
		}
	}
	
	/**
		Add, update or delete a single-valued annotation. 
		See {@link #setValues(AnnotationType, List, Unit, boolean)} for 
		more information.
	*/
	public Change setValue(AnnotationType annotationType, Object value, Unit unit, boolean merge)
	{
		return setValues(annotationType, value == null ? null : Collections.singletonList(value), unit, merge);
	}
	
	/**
		Add, update or delete annotation values for the current item.
		If the current item doesn't have any values for the annotation type
		the outcome is one of the following:
		
		<ul>
		<li>ADDED: if there is at least one value in the list
		<li>NO_CHANGE: if the list of values is null or empty
		</ul>
		
		If the current item already have values for the annotation type
		the outcome is one of the following:
		
		<ul>
		<li>UPDATED: if the values are different from the current
		<li>DELETED: if the list of values is null or empty
		<li>NO_CHANGE: if the values are the same as the existing values
		</ul>
		
		The given list of values are of course always validated against the
		specification in the annotations (eg. correct value type, allowed by
		the enumeration, within bounds of min and max values, etc.)

		@param annotationType The annotation type, which must be one of the annotation types
			specified when creating the batcher
		@param values The values to set or null/empty to remove values
		@param unit The unit of the values. Ignored if the annotation type doesn't use units.
		@param merge If TRUE and the annotation type allows an unlimted number 
			of annotations (multiplicity=0) merges the new values with the existing values.
			Ignored if multiplicity &gt; 0.
		
		@return The resulting change
	*/
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public Change setValues(AnnotationType annotationType, List<?> values, Unit unit, boolean merge)
	{
		if (currentItem == null) throw new IllegalStateException("Must call setCurrentItem() first");
		if (annotationType == null) throw new NullPointerException("annotationType (currentItem: " + currentItem + ")");
		CurrentAnnotationInfo info = currentInfo.get(annotationType.getId());
		if (info == null)
		{
			throw new InvalidDataException("Annotation type is not among the specified annotation types: " + annotationType);
		}
		
		if (info.hasBeenUsed)
		{
			throw new IllegalStateException("Not allowed to call this method with the same annotation more than once: " + annotationType);
		}
		
		if (isDebugEnabled)
		{
			log.trace("Setting values for annotation type " + annotationType + ": " + values);
		}
		
		info.hasBeenUsed = true;
		info.newValues = values;
		
		Change changeType = Change.NO_CHANGE;
		try
		{
			if (values == null || values.size() == 0)
			{
				if (info.annotationId != null)
				{
					if (isDebugEnabled) 
					{
						log.trace("Deleting " + annotationType + " from current item " + currentItem);
					}
					
					// Delete the annotation
					changeType = Change.DELETED;
	
					// Nullify "inherited_id" in cloned annotations
					nullifyClonedReferences.addToBatch(info.annotationId);
	
					// Delete inherited annotations
					deleteInheritedAnnotations.addToBatch(info.annotationId);
					
					// Delete from Annotations
					deleteFromAnnotations.addToBatch(info.annotationId);

					// Delete from **Values
					deleteValues[info.valueType.ordinal()].addToBatch(info.valueId);
					
					// Delete from ParameterValues
					deleteFromParameterValues.addToBatch(info.valueId);
				}
			}
			else
			{
				// Check multiplicity
				int multiplicity = annotationType.getMultiplicity();
				if (multiplicity > 0 && values.size() > multiplicity)
				{
					throw new InvalidDataException("Too many values. Max "+multiplicity+" value(s) are allowed for this annotation type: "+annotationType.getName());
				}
				
				// Check that values are allowed by annotation type
				// and convert to the default unit if it is needed
				UnitConverter converter = null;
				if (info.defaultUnit != null)
				{
					// The annotation use units
					if (unit == null)
					{
						// Assume values use the default unit if none was specifed
						unit = info.defaultUnit;
					}
					else if (!unit.equals(info.defaultUnit))
					{
						// Two different units, convert the values to the default unit
						converter = getUnitConverter(unit, info.defaultUnit);
					}
					info.unitId = unit.getId();
				}
				
				List<Object> convertedValues = new ArrayList<Object>(values.size());
				for (Object value : values)
				{
					if (value instanceof Number)
					{
						Number numericValue = (Number)value;
						if (converter != null) 
						{
							numericValue = converter.convertToSpecificUnit(numericValue.doubleValue());
						}
						value = info.valueType.convertNumber(numericValue);
					}
					annotationType.validateAnnotationValue(value);
					convertedValues.add(value);
				}
				info.newValues = convertedValues;
				
				if (info.annotationId != null)
				{
					// Update the current annotation if the value(s) have changed
					boolean hasChanged = false;
					if (info.valueType.isTemporal())
					{
						List d1 = convertedValues;
						List d2 = info.values;
						hasChanged = !DateUtil.safeInvariantEquals(d1, d2, info.valueType == Type.TIMESTAMP);
					}
					else
					{
						hasChanged = !EqualsHelper.invariantEquals(convertedValues, info.values);
					}
					
					if (hasChanged)
					{
						if (isDebugEnabled) 
						{
							log.trace("Updating " + annotationType + " for current item " + currentItem);
						}
						
						// Update
						changeType = Change.UPDATED;
						
						// Delete from **Values unless merging
						if (!merge || multiplicity != 0)
						{
							deleteValues[annotationType.getValueType().ordinal()].addToBatch(info.valueId);
						}
						
						// Insert into **Values
						for (Object v : convertedValues)
						{
							insertValues[info.valueType.ordinal()].addToBatch(info.valueId, v);
						}
						
						// Update last_update, unit, version in Annotations
						updateAnnotations.addToBatch(info.version+1, info.unitId, batchDate, info.annotationId, info.version);
					}
					else
					{
						if (isDebugEnabled)
						{
							log.trace("No change " + annotationType + " for current item " + currentItem);
						}
					}
				}
				else
				{
					if (isDebugEnabled) 
					{
						log.trace("Creating " + annotationType + " for current item " + currentItem);
					}
					
					// Create a new annotation
					changeType = Change.ADDED;

					// Insert into ParameterValues
					info.valueId = insertIntoParameterValues.insert(discriminators[info.valueType.ordinal()]);
					
					// Insert into **Values
					for (Object v : convertedValues)
					{
						insertValues[info.valueType.ordinal()].addToBatch(info.valueId, v);
					}

					if (currentAnnotationSetId == 0)
					{
						// Insert into AnnotationSet
						currentAnnotationSetId = insertIntoAnnotationSets.insert(currentItem.getId());

						// Update item table, annotationset_id column
						updateItemTable.addToBatch(currentAnnotationSetId, currentItem.getId());
						if (evictFromSecondLevelCacheIfModified)
						{
							// We must evict the entity from the Hibernate second-level 
							// cache or Hibernate may use the old data
							secondLevelCache.evictEntity(itemType.getDataClass(), currentItem.getId());
						}
					}

					// Insert into Annotations
					info.annotationId = insertIntoAnnotations.insert(currentAnnotationSetId, annotationType.getId(), info.unitId, info.valueId, batchDate);
				}
			}
		}
		catch (SQLException ex)
		{
			throw new BaseException(ex);
		}
		
		// Store the annotation set ID so that we can tell
		// the SnapshotManager to clear the cache when this
		// batcher is closed
		if (changeType != Change.NO_CHANGE) 
		{
			if (deleteSnapshotIfModifed)
			{
				deleteSnapshotIfModifed = false;
				snapshotsToDelete.add(currentAnnotationSetId);
			}
			if (loggingInterceptor != null)
			{
				loggingInterceptor.addManualSingleChange(changeType.getLogChangeType(), info);
			}
		}
		
		return changeType;
	}
	
	/**
		A flag indicating what type of change that was made to an
		annotation as a result of calling
		{@link AnnotationBatcher#setValues(AnnotationType, List, Unit, boolean)}
	*/
	public static enum Change
	{
		/**
			The annotation was not changed.
		*/
		NO_CHANGE(null),
		
		/**
			A new annotation was created.
		*/
		ADDED(ChangeType.ADD),
		
		/**
			An existing annotation was updated.
		*/
		UPDATED(ChangeType.UPDATE),
		
		/**
			An existing annotation was deleted.
		*/
		DELETED(ChangeType.REMOVE);
		
		private final ChangeType logChangeType;
		
		Change(ChangeType logChangeType)
		{
			this.logChangeType = logChangeType;
		}
		
		ChangeType getLogChangeType()
		{
			return logChangeType;
		}
	}
	
	/**
		Holds information about an annotation for the current item.
		We pre-initialize one instance for each registered annotation 
		type. In the setCurrentItem() method we update the instances
		with information about current annotations. Before using an
		entry we must always check if itemId==currentItem.id. If the
		IDs don't match the entry is for a previous item and the
		current item doesn't have any annotation for the annotation type.
		The 'hasBeenUsed' flag is set to prevent calling the setValues() method
		more than once for each annotation type.
		
		This class exposes some information publicly since it is used for
		change history logging. See {@link AnnotationLogger}.
	*/
	public static class CurrentAnnotationInfo
		implements LoggableData
	{
		// Pre-initialized
		final AnnotationType annotationType;
		final Type valueType;
		final Unit defaultUnit;
		final Item itemType;
		
		// Updated in setCurrentItem
		int itemId;
		boolean hasBeenUsed;
		Integer annotationId;
		int version;
		Integer unitId;
		int valueId;
		Date lastUpdated;
		final List<Object> values;
		List<?> newValues;
		
		CurrentAnnotationInfo(AnnotationType at, Item itemType)
		{
			this.annotationType = at;
			this.itemType = itemType;
			this.valueType = at.getValueType();
			this.defaultUnit = at.getDefaultUnit();
			this.values = new ArrayList<Object>();
		}
		
		/**
			Get the annotation type.
		*/
		public AnnotationType getAnnotationType()
		{
			return annotationType;
		}
		
		/**
			Shortcut to {@link AnnotationType#getValueType()}.
		*/
		public Type getValueType()
		{
			return valueType;
		}
		
		/**
			Get the default unit for the annotation type.
			@see AnnotationType#getDefaultUnit()
		*/
		public Unit getDefaultUnit()
		{
			return defaultUnit;
		}
		
		/**
			Get the item type of the item.
		*/
		public Item getItemType()
		{
			return itemType;
		}
		
		/**
			Get the id of the current item.
		*/
		public int getItemId()
		{
			return itemId;
		}
		
		/**
			Get the ID of the unit this annotation is using.
		*/
		public Integer getUnitId()
		{
			return unitId;
		}
		
		/**
			Get the old annotation values. If the annotation
			uses units, the values are always in the default unit
			specified by the annotation type.
		*/
		public List<?> getOldValues()
		{
			return values;
		}
		
		/**
			Get the new annotation values. If the annotation
			uses units, the values are always in the default unit
			specified by the annotation type.
		*/
		public List<?> getNewValues()
		{
			return newValues;
		}
	}
	
	/**
		Batcher implementation for inserting into tables.
		Assumes that the table has an "id" column that is
		auto-generated by the SQL statement. The value of
		the generated id is retuned by the {@link #insert(Object[])}
		method.
	*/
	static class InsertBatcher
	{
		private final String sql;
		private final PreparedStatement statement;
		private final int[] parameterTypes;
		
		/**
			Prepare a new batcher for the given SQL statement.
			@param sql A SQL statement
			@param parameterTypes Parameter type codes as defined in {@link Types}.
		*/
		InsertBatcher(Connection c, String sql, int... parameterTypes)
			throws SQLException
		{
			log.debug(sql);
			this.sql = sql;
			this.statement = c.prepareStatement(sql, new String[] { "id" } );
			this.parameterTypes = parameterTypes;
		}
		
		/**
			The SQL statement this batcher is executing.
		*/
		String getSql()
		{
			return sql;
		}

		/**
			Execute an insert SQL statement with the given parameter
			values. The number of values must match the number of 
			parameter types. 
			@return The ID of inserted row
		*/
		int insert(Object... values)
			throws SQLException
		{
			if (values.length != parameterTypes.length)
			{
				throw new SQLException("Number of values (" + values.length + 
					") doesn't match the number of defined parameters (" + parameterTypes.length + ") in SQL: " + sql);
			}
			int i = 0;
			for (Object v : values)
			{
				int type = parameterTypes[i];
				statement.setObject(++i, v, type);
			}
				
			int num = statement.executeUpdate();
			ResultSet keys = statement.getGeneratedKeys();
			keys.next();
			return keys.getInt(1);
		}
		
		/**
			Close the batcher.
		*/
		void close()
		{
			try
			{
				statement.close();
			}
			catch (SQLException ex)
			{
				log.error(ex.getMessage(), ex);
				LogUtil.logSQLExceptionChain(log, ex);
			}
		}

	}
	
	/**
		Batcher implementation for updating and deleting from tables.
	*/
	static class UpdateBatcher
	{
		private final String sql;
		private final PreparedStatement statement;
		private final int[] parameterTypes;
		private int count;
		
		/**
			Prepare a new batcher for the given SQL statement.
			@param sql A SQL statement
			@param parameterTypes Parameter type codes as defined in {@link Types}.
		*/
		UpdateBatcher(Connection c, String sql, int... parameterTypes)
			throws SQLException
		{
			log.debug(sql);
			this.sql = sql;
			this.statement = c.prepareStatement(sql);
			this.parameterTypes = parameterTypes;
		}
		
		/**
			The SQL statement this batcher is executing.
		*/
		String getSql()
		{
			return sql;
		}
		
		/**
			Add a new entry to the batch. The number of values must match the number of 
			parameter types. 
		*/
		void addToBatch(Object... values)
			throws SQLException
		{
			if (values.length != parameterTypes.length)
			{
				throw new SQLException("Number of values (" + values.length + 
					") doesn't match the number of defined parameters (" + parameterTypes.length + ") in SQL: " + sql);
			}
			int i = 0;
			for (Object v : values)
			{
				int type = parameterTypes[i];
				statement.setObject(++i, v, type);
			}
			statement.addBatch();
			count++;
		}
		
		/**
			Flush and execute the statements added to the batch.
		*/
		void flush()
			throws SQLException
		{
			if (count > 0)
			{
				statement.executeBatch();
				count = 0;
			}
		}

		/**
			Close the batcher.
		*/
		void close()
		{
			try
			{
				statement.close();
			}
			catch (SQLException ex)
			{
				log.error(ex.getMessage(), ex);
				LogUtil.logSQLExceptionChain(log, ex);
			}
		}
	}
	
}
