/*
	$Id: MultiPermissions.java 7182 2016-06-23 11:53:54Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.core;

import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.data.GroupData;
import net.sf.basedb.core.data.ProjectData;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.EqualsHelper;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Expression;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;

/**
	This is a helper class that is very useful when you have a collection of
	{@link Shareable} items and want to modify some permissions on
	all of them without affecting the rest of the permissions. 
	<p>
	For example it is possible to add write permission for a new user
	while preserving the permissions already given to other user, groups
	and projects. This works even if the logged in user doesn't have read
	permission to all other user, groups and projects the items have been
	shared to.
	<p>
	It is a four-step procedure:
	<ul>
	<li>Create a new object using the {@link #MultiPermissions(Collection)}
		constructor.
	<li>Add, remove or modify permissions as required.
	<li>Call the {@link #updateKeys(DbControl)} method to generate the new keys 
		and assign those to the items
	<li>Commit the transaction.
	</ul>
	@author Nicklas
	@version 2.0
	@base.modified $Date: 2016-06-23 13:53:54 +0200 (to, 23 jun 2016) $
*/
public class MultiPermissions
{

	/**
		Check if a collection of <code>Set</code>:s all contain
		the same permissions.
		@param permissions The collection of sets to check
		@return TRUE If all sets in the collection contains the same
			permissions
		@see Permission
	*/
	public static boolean allSame(Collection<Set<Permission>> permissions)
	{
		Set<Permission> first = null;
		for (Set<Permission> p : permissions)
		{
			if (first == null)
			{
				first = p;
			}
			else
			{
				if (!first.equals(p)) return false;
			}
		}
		return true;
	}

	private final Set<SharedItem> items;
	private final Map<ItemKey, UserPermissions> userPermissions;
	private final Map<ItemKey, GroupPermissions> groupPermissions;
	private final Map<ProjectKey, ProjectPermissions> projectPermissions;

	/**
		Create a new <code>MultiPermissions</code> object and feed it the
		collection of {@link Shareable} items that you want to modify.
		@param itemsToShare The collection of items
		@throws InvalidDataException If the collection is null or empty
		@throws BaseException If there is another error
	*/
	public MultiPermissions(Collection<? extends SharedItem> itemsToShare)
		throws InvalidDataException, BaseException
	{
		if (itemsToShare == null || itemsToShare.size() == 0) throw new InvalidUseOfNullException("items");
		this.items = new HashSet<SharedItem>(itemsToShare.size());
		this.userPermissions = new HashMap<ItemKey, UserPermissions>();
		this.groupPermissions = new HashMap<ItemKey, GroupPermissions>();
		this.projectPermissions = new HashMap<ProjectKey, ProjectPermissions>();
		for (SharedItem item : itemsToShare)
		{
			if (items.add(item))
			{
				ItemKey itemKey = item.getItemKey();
				ProjectKey projectKey = item.getProjectKey();
				if (!userPermissions.containsKey(itemKey))
				{
					userPermissions.put(itemKey, itemKey == null ? new UserPermissions() : new UserPermissions(itemKey));
					groupPermissions.put(itemKey, itemKey == null ? new GroupPermissions() : new GroupPermissions(itemKey));
				}
				if (!projectPermissions.containsKey(projectKey))
				{
					projectPermissions.put(projectKey, projectKey == null ? new ProjectPermissions() : new ProjectPermissions(projectKey));
				}
			}
		}
	}

	/**
		Grant permissions to a user. Use an empty <code>Set</code> or null
		to remove all permissions for the user.

		@param user The <code>User</code>
		@param permissions The permissions to grant
		@throws InvalidDataException If the user is null
		@see Permission
	*/
	public void setPermissions(User user, Set<Permission> permissions)
		throws InvalidDataException
	{
		if (user == null) throw new InvalidUseOfNullException("user");
		for (UserPermissions up : userPermissions.values())
		{
			up.setPermissions(user, permissions);
		}
	}

	/**
		Grant permissions to a user, keeping those that have already been set.
		
		@param user The <code>User</code>
		@param permissions The permissions to grant
		@throws InvalidDataException If the user is null
		@see Permission
	*/
	public void addPermissions(User user, Set<Permission> permissions)
		throws InvalidDataException
	{
		if (user == null) throw new InvalidUseOfNullException("user");
		if (permissions != null && !permissions.isEmpty())
		{
			for (UserPermissions up : userPermissions.values())
			{
				up.addPermissions(user, permissions);
			}
		}
	}

	/**
		Get the permissions for a user. The returned map contains one entry
		for each unique {@link ItemKey} found among the items passed
		to the constructor, including the null item key if present. The value
		of each entry is a <code>Set</code> containing the granted permissions.
		
		@param user The <code>User</code> for which we want to get the permission
		@return A <code>Map</code> containing a <code>Set</code> of
			the granted permissions for each <code>ItemKey</code>
		@throws InvalidDataException If the user is null
		@see Permission
	*/
	public Map<ItemKey, Set<Permission>> getPermissions(User user)
		throws InvalidDataException
	{
		if (user == null) throw new InvalidUseOfNullException("user");
		Map<ItemKey, Set<Permission>> permissions = new HashMap<ItemKey, Set<Permission>>(userPermissions.size());
		for (Map.Entry<ItemKey, UserPermissions> entry : userPermissions.entrySet())
		{
			permissions.put(entry.getKey(), entry.getValue().getPermissions(user));
		}
		return permissions;
	}
	
	/**
		Grant permissions to a group. Use an empty <code>Set</code> or null
		to remove all permissions for the group.

		@param group The <code>Group</code>
		@param permissions The permissions to grant
		@throws InvalidDataException If the group is null
		@see Permission
	*/
	public void setPermissions(Group group, Set<Permission> permissions)
		throws InvalidDataException
	{
		if (group == null) throw new InvalidUseOfNullException("group");
		for (GroupPermissions gp : groupPermissions.values())
		{
			gp.setPermissions(group, permissions);
		}
	}

	/**
		Grant permissions to a group, keeping those that have already been set.
		
		@param group The <code>Group</code>
		@param permissions The permissions to grant
		@throws InvalidDataException If the group is null
		@see Permission
	*/
	public void addPermissions(Group group, Set<Permission> permissions)
		throws InvalidDataException
	{
		if (group == null) throw new InvalidUseOfNullException("group");
		if (permissions != null && !permissions.isEmpty())
		{
			for (GroupPermissions gp : groupPermissions.values())
			{
				gp.addPermissions(group, permissions);
			}
		}
	}

	/**
		Get the permissions for a group. The returned map contains one entry
		for each unique {@link ItemKey} found among the items passed
		to the constructor, including the null item key if present. The value
		of each entry is a <code>Set</code> containing the granted permissions.
		
		@param group The <code>Group</code> for which we want to get the permission
		@return A <code>Map</code> containing a <code>Set</code> of
			the granted permissions for each <code>ItemKey</code>
		@throws InvalidDataException If the group is null
		@see Permission
	*/
	public Map<ItemKey, Set<Permission>> getPermissions(Group group)
		throws InvalidDataException
	{
		if (group == null) throw new InvalidUseOfNullException("group");
		Map<ItemKey, Set<Permission>> permissions = new HashMap<ItemKey, Set<Permission>>(groupPermissions.size());
		for (Map.Entry<ItemKey, GroupPermissions> entry : groupPermissions.entrySet())
		{
			permissions.put(entry.getKey(), entry.getValue().getPermissions(group));
		}
		return permissions;
	}

	/**
		Grant permissions to a project. Use an empty <code>Set</code> or null
		to remove all permissions for the project.

		@param project The <code>Project</code>
		@param permissions The permissions to grant
		@throws InvalidDataException If the project is null
		@throws PermissionDeniedException If the logged in user's doesn't have 
			{@link Permission#USE} permission to the project
		@see Permission
	*/
	public void setPermissions(Project project, Set<Permission> permissions)
		throws PermissionDeniedException, InvalidDataException
	{
		if (project == null) throw new InvalidUseOfNullException("project");
		for (ProjectPermissions pp : projectPermissions.values())
		{
			pp.setPermissions(project, permissions);
		}
	}

	/**
		Grant permissions to a project, keeping those that have already been set.
		
		@param project The <code>Project</code>
		@param permissions The permissions to grant
		@throws InvalidDataException If the project is null
		@throws PermissionDeniedException If the logged in user's doesn't have 
			{@link Permission#USE} permission to the project
		@see Permission
	*/
	public void addPermissions(Project project, Set<Permission> permissions)
		throws PermissionDeniedException, InvalidDataException
	{
		if (project == null) throw new InvalidUseOfNullException("project");
		if (permissions != null && !permissions.isEmpty())
		{
			for (ProjectPermissions pp : projectPermissions.values())
			{
				pp.addPermissions(project, permissions);
			}
		}
	}

	/**
		Get the permissions for a project. The returned map contains one entry
		for each unique {@link ProjectKey} found among the items passed
		to the constructor, including the null project key if present. The value
		of each entry is a <code>Set</code> containing the granted permissions.
		
		@param project The <code>Project</code> for which we want to get the permission
		@return A <code>Map</code> containing a <code>Set</code> of
			the granted permissions for each <code>ProjectKey</code>
		@throws InvalidDataException If the project is null
		@see Permission
	*/
	public Map<ProjectKey, Set<Permission>> getPermissions(Project project)
		throws InvalidDataException
	{
		if (project == null) throw new InvalidUseOfNullException("project");
		Map<ProjectKey, Set<Permission>> permissions = new HashMap<ProjectKey, Set<Permission>>(projectPermissions.size());
		for (Map.Entry<ProjectKey, ProjectPermissions> entry : projectPermissions.entrySet())
		{
			permissions.put(entry.getKey(), entry.getValue().getPermissions(project));
		}
		return permissions;
	}
	
	/**
		Merge the permissions from the given item key keeping those
		that has already been set.

		@param itemKey An item key (if null, this method simply returns)
		@since 2.16
		@see UserPermissions#merge(ItemKey)
		@see GroupPermissions#merge(ItemKey)
	*/
	public void merge(ItemKey itemKey)
	{
		if (itemKey == null) return;
		for (UserPermissions up : userPermissions.values())
		{
			up.merge(itemKey);
		}
		for (GroupPermissions gp : groupPermissions.values())
		{
			gp.merge(itemKey);
		}
	}
	
	/**
		Merge the permissions from the given project key keeping those
		that has already been set.
	
		@param projectKey An project key (if null, this method simply returns)
		@since 2.16
		@see ProjectPermissions#merge(ProjectKey)
	*/
	public void merge(ProjectKey projectKey)
	{
		if (projectKey == null) return;
		for (ProjectPermissions pp : projectPermissions.values())
		{
			pp.merge(projectKey);
		}
	}
	
	/**
		Get a query that returns all users appearing in at least one of
		the item keys. The query will not return users that the logged in 
		user doesn't have permission to read. 
	*/
	public ItemQuery<User> getUsers()
	{
		Set<Integer> userIds = new HashSet<Integer>();
		for (UserPermissions up : userPermissions.values())
		{
			Set<UserData> users = up.getUserPermissions().keySet();
			for (UserData u : users)
			{
				userIds.add(u.getId());
			}
		} 
		ItemQuery<User> query = User.getQuery();
		query.include(Include.ALL);
		if (userIds.size() > 0)
		{
			query.restrict(
				Restrictions.in(
					Hql.property("id"),
					Expressions.parameter("userids")
				)
			);
			query.setParameter("userids", userIds, Type.INT);
		}
		else
		{
			Expression id = Hql.property("id");
			query.restrict(Restrictions.neq(id, id));
		}
		return query;
	}
	
	/**
		Get a query that returns all groups appearing in at least one of
		the item keys. The query will not return groups that the logged in 
		user doesn't have permission to read. 
	*/
	public ItemQuery<Group> getGroups()
	{
		Set<Integer> groupIds = new HashSet<Integer>();
		for (GroupPermissions gp : groupPermissions.values())
		{
			Set<GroupData> groups = gp.getGroupPermissions().keySet();
			for (GroupData g : groups)
			{
				groupIds.add(g.getId());
			}
		} 
		ItemQuery<Group> query = Group.getQuery();
		query.include(Include.ALL);
		if (groupIds.size() > 0)
		{
			query.restrict(
				Restrictions.in(
					Hql.property("id"),
					Expressions.parameter("groupids")
				)
			);
			query.setParameter("groupids", groupIds, Type.INT);
		}
		else
		{
			Expression id = Hql.property("id");
			query.restrict(Restrictions.neq(id, id));
		}
		return query;
	}
	
	/**
		Get a query that returns all projects appearing in at least one of
		the project keys. The query will not return projects that the logged in 
		user doesn't have permission to read. 
	*/
	public ItemQuery<Project> getProjects()
	{
		Set<Integer> projectIds = new HashSet<Integer>();
		for (ProjectPermissions pp : projectPermissions.values())
		{
			Set<ProjectData> projects = pp.getProjectPermissions().keySet();
			for (ProjectData p : projects)
			{
				projectIds.add(p.getId());
			}
		} 
		ItemQuery<Project> query = Project.getQuery();
		query.include(Include.ALL);
		if (projectIds.size() > 0)
		{
			query.restrict(
				Restrictions.in(
					Hql.property("id"),
					Expressions.parameter("projectids")
				)
			);
			query.setParameter("projectids", projectIds, Type.INT);
		}
		else
		{
			Expression id = Hql.property("id");
			query.restrict(Restrictions.neq(id, id));
		}
		return query;
	}
	
	/**
	 	Returns the items that are handled within this class. 
	*/
	public Set<SharedItem> getItems()
	{	
		return items;
	}
	
	/**
		Get all items of a specific item type. The returned set
		is never null, but may be empty if no items of the specified
		type was found.
		
		@param type The item type to search for
		@return A set with the items
		@since 2.8
	*/
	public Set<SharedItem> getItems(Item type)
	{
		return getItems(Collections.singleton(type));
	}
	
	/**
		Get all items of specific item types. The returned set
		is never null, but may be empty if no items of the specified
		types was found.
		
		@param types A collection with the item types 
			to search for
		@return A set with the items
		@since 2.8
	*/
	public Set<SharedItem> getItems(Collection<Item> types)
	{
		Set<SharedItem> found = new HashSet<SharedItem>();
		for (SharedItem item : items)
		{
			if (types.contains(item.getType())) found.add(item);
		}
		return found;
	}
	
	/**
		Generate the new keys and update all items with the new keys.
		@param dc The {@link DbControl} object to use for database access
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#SET_PERMISSION} permission on all items
		@throws BaseException If there is another error
	*/
	public void updateKeys(DbControl dc)
		throws PermissionDeniedException, BaseException
	{
		updateKeys(dc, false);
	}

	/**
		Generate the new keys and update all items with the new keys.
		@param dc The {@link DbControl} object to use for database access
		@param recursive If TRUE, files and subdirectories to a selected
			directory will also shared with the same permissions as the
			parent directory
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#SET_PERMISSION} permission on all items
		@throws BaseException If there is another error
		@since 2.8
	*/
	public void updateKeys(DbControl dc, boolean recursive)
		throws PermissionDeniedException, BaseException
	{
		Map<ItemKey, ItemKey> newItemKeys = new HashMap<ItemKey, ItemKey>(userPermissions.size());
		Map<ProjectKey, ProjectKey> newProjectKeys = new HashMap<ProjectKey, ProjectKey>(projectPermissions.size());
		
		for (Map.Entry<ItemKey, UserPermissions> entry : userPermissions.entrySet())
		{
			ItemKey oldItemKey = entry.getKey();
			UserPermissions up = entry.getValue();
			GroupPermissions gp = groupPermissions.get(oldItemKey);
			ItemKey newItemKey = up.size() == 0 && gp.size() == 0 ? null : ItemKey.getNewOrExisting(dc, up, gp);
			newItemKeys.put(oldItemKey, newItemKey);
		}
		
		for (Map.Entry<ProjectKey, ProjectPermissions> entry : projectPermissions.entrySet())
		{
			ProjectKey oldProjectKey = entry.getKey();
			ProjectPermissions pp = entry.getValue();
			ProjectKey newProjectKey = pp.size() == 0 ? null : ProjectKey.getNewOrExisting(dc, pp);
			newProjectKeys.put(oldProjectKey, newProjectKey);
		}
	
		for (SharedItem item : items)
		{
			item.checkPermission(Permission.SET_PERMISSION);
			dc.reattachItem(item, false);
			
			// Update keys if they have changed
			ItemKey existingItemKey = item.getItemKey();
			ItemKey newItemKey = newItemKeys.get(existingItemKey);
			if (!EqualsHelper.equals(newItemKey, existingItemKey))
			{
				item.setItemKey(newItemKey);	
			}
			
			ProjectKey existingProjectKey = item.getProjectKey();
			ProjectKey newProjectKey = newProjectKeys.get(existingProjectKey);
			if (!EqualsHelper.equals(newProjectKey, existingProjectKey))
			{
				item.setProjectKey(newProjectKey);	
			}
			
			if (recursive && item instanceof Directory)
			{
				ShareableUtil.shareRecursively(dc, (Directory)item);
			}
		}
	}
	
}
