/**
	$Id: PostgresDbEngine.java 7075 2016-02-15 10:58:10Z nicklas $

	Copyright (C) 2006, 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.dbengine;

import java.util.Collection;
import java.util.Set;

import org.hibernate.mapping.PersistentClass;

/**
	Database engine for Postgres.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2016-02-15 11:58:10 +0100 (må, 15 feb 2016) $
*/
public class PostgresDbEngine
	extends AbstractDbEngine
	implements DbEngine
{

	/**
		Create PostgresDbEngine.
	*/
	public PostgresDbEngine()
	{}

	/*
		From the DbEngine interface
		-------------------------------------------
	*/
	/**
		Always false.
	*/
	@Override
	public boolean useThetaJoin()
	{
		return false;
	}
	
	/**
		Generate <code>create schema &lt;schema&gt;</code>.
		The catalog parameter is not supported.
		@since 3.4
	*/
	@Override
	public String getCreateSchemaSql(String catalog, String schema)
	{
		StringBuilder sql = new StringBuilder();
		sql.append("create schema ").append(getQuotedName(schema));
		return sql.toString();
	}
	
	/**
		Generate <code>alter table &lt;schema&gt;.&lt;table&gt; add constraint &lt;name&gt;
		unique (&lt;columns&gt;)</code> for a unique index, 
		<code>create index &lt;name&gt; on &lt;schema&gt;.&lt;table&gt; (&lt;columns&gt;)</code>
		for other indexes.
		The catalog parameter is not supported.
	*/
	@Override
	public String getCreateIndexSql(String catalog, String schema, String table, String name, 
		Set<String> columns, boolean unique)
	{
		StringBuilder sql = new StringBuilder();
		if (unique)
		{
			sql.append("alter table ");
			if (schema != null) sql.append(getQuotedName(schema)).append(".");
			sql.append(getQuotedName(table)).append(" add constraint ");
			sql.append(getQuotedName(name)).append(" unique (");
			appendList(sql, columns);
			sql.append(")");
		}
		else
		{
			sql.append("create index ");
			sql.append(getQuotedName(name)).append(" on ");
			if (schema != null) sql.append(getQuotedName(schema)).append(".");
			sql.append(getQuotedName(table)).append(" (");
			appendList(sql, columns);
			sql.append(")");
			
		}
		return sql.toString();
	}
	
	/**
		Generate <code>alter table &lt;schema&gt;.&lt;table&gt; drop constraint &lt;name&gt;</code>
		for unique indexes, <code>drop index &lt;schema&gt;.&lt;name&gt;</code> for other indexes.
		The catalog parameter is not supported.
	*/
	@Override
	public String getDropIndexSql(String catalog, String schema, String table, String name, boolean unique)
	{
		StringBuilder sql = new StringBuilder();
		if (unique)
		{
			sql.append("alter table ");
			if (schema != null) sql.append(getQuotedName(schema)).append(".");
			sql.append(getQuotedName(table)).append(" drop constraint ");
			sql.append(getQuotedName(name));
		}
		else
		{
			sql.append("drop index ");
			if (schema != null) sql.append(getQuotedName(schema)).append(".");
			sql.append(getQuotedName(name));
		}
		return sql.toString();
	}

	/**
		Generate <code>alter table schema.table add constraint name
		foreign key (columns) references refTable (refColumns)</code>
		@since 3.1
	 */
	@Override
	public String getCreateForeignKeySql(String catalog, String schema,	String table, 
		String name, Set<String> columns, String refTable, Set<String> refColumns) 
	{
		StringBuilder sql = new StringBuilder();
		
		sql.append("alter table ");
		if (schema != null) sql.append(getQuotedName(schema)).append(".");
		sql.append(getQuotedName(table));
		sql.append(" add constraint ");
		sql.append(getQuotedName(name));
		sql.append(" foreign key (");
		appendList(sql, columns);
		sql.append(") references ");
		sql.append(getQuotedName(refTable));
		sql.append(" (");
		appendList(sql, refColumns);
		sql.append(")");
		return sql.toString();
	}

	/**
		Generate <code>alter table &lt;schema&gt;.&lt;table&gt; DROP CONSTRAINT &lt;name&gt;</code>
		@since 3.0
	*/
	@Override
	public String getDropForeignKeySql(String catalog, String schema, String table, String name)
	{
		return getDropIndexSql(catalog, schema, table, name, true);
	}
	
	
	/**
		Generate <code>alter table schema.table add constraint name
		primary key (columns)</code>.
		@since 3.1
	*/
	@Override
	public String getCreatePrimaryKeySql(String catalog, String schema, String table, String name, 
		Set<String> columns)
	{
		StringBuilder sql = new StringBuilder();
		sql.append("alter table ");
		if (schema != null) sql.append(getQuotedName(schema)).append(".");
		sql.append(getQuotedName(table)).append(" add constraint ");
		sql.append(getQuotedName(name)).append(" primary key (");
		appendList(sql, columns);
		sql.append(")");
		return sql.toString();
	}

	/**
		Generate <code>alter table schema.table drop constraint name</code>
		@since 3.1
	*/
	@Override
	public String getDropPrimaryKey(String catalog, String schema, String table, String name) 
	{
		return getDropIndexSql(catalog, schema, table, name, true);
	}

	/**
		@return Always "nextval('sequenceName')" (assuming that the ID column is generated by a sequence)
		@since 3.0
	*/
	@Override
	public String getInsertAutoIncrementSql(PersistentClass pc)
	{
		// For sequence generators we must use: 
		// INSERT INTO ... (id, ...) VALUES (nextval('sequenceName'), ...)
			
		// The sequence name is found in the pc.identifierProperty.identifierGeneratorProperties
		org.hibernate.mapping.SimpleValue idValue = (org.hibernate.mapping.SimpleValue)pc.getIdentifierProperty().getValue();
		String sequenceName = idValue.getIdentifierGeneratorProperties().getProperty(org.hibernate.id.enhanced.SequenceStyleGenerator.SEQUENCE_PARAM, "hibernate_sequence");
		return "nextval('" + sequenceName + "')";
	}

	/**
		Return true. Otherwise queries may take a very long time since
		indexes may not be used properly.
	*/
	@Override
	public boolean analyzeAfterBatchOperation()
	{
		return true;
	}

	/**
		Generate <code>VACUUM FULL &lt;schema&gt;.&lt;table&gt;</code>.
		The catalog parameter is not supported.
	*/
	@Override
	public String getOptimizeTableSql(String catalog, String schema, String table)
	{
		schema = schema == null ? "" : getQuotedName(schema) + ".";
		return "VACUUM FULL " + schema + getQuotedName(table);
	}
	
	/**
		Generate <code>ANALYZE &lt;schema&gt;.&lt;table&gt;</code>.
		The catalog parameter is not supported.
	*/
	@Override
	public String getAnalyzeTableSql(String catalog, String schema, String table)
	{
		schema = schema == null ? "" : getQuotedName(schema) + ".";
		return "ANALYZE " + schema + getQuotedName(table);
	}
	/**
		Returns FALSE.
		@since 2.4
	*/
	@Override
	public boolean caseInsensitiveComparison()
	{
		return false;
	}
	/**
		Returns FALSE.
	*/
	@Override
	public boolean supportsColumnAliasInWhere()
	{
		return false;
	}
	/**
		Returns TRUE.
	*/
	@Override
	public boolean supportColumnAliasInOrderBy()
	{
		return true;
	}
	/**
		Returns TRUE.
	*/
	@Override
	public boolean supportColumnAliasInGroupBy()
	{
		return true;
	}
	/**
		Returns FALSE.
	*/
	@Override
	public boolean supportColumnAliasInHaving()
	{
		return false;
	}
	/**
		Returns TRUE.
	*/
	@Override
	public boolean checkForInvalidNumberOperation()
	{
		return true;
	}
	/**
		Returns TRUE.
	*/
	@Override
	public boolean selectOrderByColumnsIfDistinct()
	{
		return true;
	}
	/**
		Return <code>&lt;value&gt; ~ &lt;regexp&gt;</code>.
		@since 2.8
	*/
	@Override
	public String rlike(String value, String regexp)
	{
		return value + " ~ " + regexp;
	}

	/**
		@return Always TRUE
	 */
	@Override
	public boolean useSavePointToContinueTransactionFromSqlFailure() 
	{
		return true;
	}
	
	@Override
	public String getApproximateRowCountSql(String catalog, String schema, String table)
	{
		return "select reltuples from pg_class where relname = '" + table + "'";
	}
	// -------------------------------------------


	private void appendList(StringBuilder sb, Collection<String> items)
	{
		int i = 0;
		for (String column : items)
		{
			if (i > 0) sb.append(", ");
			sb.append(getQuotedName(column));
			i++;
		}
	}
	

}
