/**
	$Id: TableInfo.java 7075 2016-02-15 10:58:10Z nicklas $

	Copyright (C) 2006 Johan Enell, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.
	
	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.dbengine;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.hibernate.dialect.Dialect;
import org.hibernate.mapping.Column;
import org.hibernate.mapping.ForeignKey;
import org.hibernate.mapping.Index;
import org.hibernate.mapping.PrimaryKey;
import org.hibernate.mapping.Table;
import org.hibernate.mapping.UniqueKey;

/**
	Holds minimal but useful information about a table in the database.
	The information can be populated either from Hibernate mappings or
	from JDBC metadata about the database. This makes it possible to compare
	the structure of the current database with the information in the Hibernate
	mapping documents.
	<p>
	The main reason for this class to exists is that we want to
	solve the problem Hibernate has with creating indexes when doing an
	update. The information in this class makes it possible to find which indexes
	are missing. As a bonus we also have code to drop and re-create all indexes.
	
	@author nicklas
	@version 2.0
	@base.modified $Date: 2016-02-15 11:58:10 +0100 (må, 15 feb 2016) $
*/
public class TableInfo
{

	/**
		List all tables that are present in the given catalog and/or schema.
		@param metaData A metadata object that is connected to a database
		@param catalog The name of the catalog, or null
		@param schema The name of the schema, or null
		@return A list of tables
		@since 3.1
	*/
	public static List<Table> listTables(DatabaseMetaData metaData, String catalog, String schema)
		throws SQLException
	{
		List<Table> tables = new LinkedList<Table>();
		ResultSet result = metaData.getTables(catalog, schema, null, new String[] { "TABLE" });
		while (result.next())
		{
			String tableName = result.getString("TABLE_NAME");
			String tableCatalog = result.getString("TABLE_CAT");
			String tableSchema = result.getString("TABLE_SCHEM");
			Table table = new Table();
			table.setName(tableName);
			table.setCatalog(tableCatalog);
			table.setSchema(tableSchema);
			tables.add(table);
		}
		return tables;
	}
	
	private final Table table;
	private Dialect dialect;
	private boolean existsInDb;
	
	private Set<ColumnInfo> columns;
	private PrimaryKeyInfo pkInfo;
	private Set<ForeignKeyInfo> foreignKeys;
	private Set<IndexInfo> indexes;
	private long numberOfRows = -1;
	private long approximateNumberOfRows = -1;
	
	
	
	private TableInfo(Table table)
	{
		this.table = table;
		columns = new LinkedHashSet<ColumnInfo>();
		foreignKeys = new LinkedHashSet<ForeignKeyInfo>();
		indexes = new LinkedHashSet<IndexInfo>();
	}
	
	
	/**
		Create table info by reading JDBC metadata from the current database
		
		@param table The Hibernate table object
		@param metaData The JDBC metadata connected to the current database
		@throws SQLException If there is an error
	*/
	public TableInfo(Table table, DatabaseMetaData metaData)
			throws SQLException
	{
		this(table, metaData, null, null);
	}

	/**
		Create table info by reading JDBC metadata from the current database
		
		@param table The Hibernate table object
		@param metaData The JDBC metadata connected to the current database
		@param defaultCatalog Catalog to use if the table doesn't specify a catalog
		@param defaultSchema Schema to use if the table doesn't specify a catalog
		@throws SQLException If there is an error
		@since 3.4
	*/
	public TableInfo(Table table, DatabaseMetaData metaData, String defaultCatalog, String defaultSchema)
		throws SQLException
	{
		this(table);
		
		String catalog = table.getCatalog();
		if (catalog == null) catalog = defaultCatalog;
		String schema = table.getSchema();
		if (schema == null) schema = defaultSchema;
		
		ResultSet tables = metaData.getTables(catalog, schema, table.getName(), null);
		existsInDb = tables.next();
		tables.close();
		if (!existsInDb) return;
		
		// Load column information
		ResultSet sqlColumns = metaData.getColumns(catalog, schema, table.getName(), "%"); 
		while (sqlColumns.next())
		{
			String name = makeSafe(sqlColumns.getString("COLUMN_NAME"));
			String type = sqlColumns.getString("TYPE_NAME");
			int sqlType = sqlColumns.getInt("DATA_TYPE");
			int size = sqlColumns.getInt("COLUMN_SIZE");
			boolean isNullable = sqlColumns.getInt("NULLABLE") == DatabaseMetaData.columnNullable;
			columns.add(new ColumnInfo(name, type, sqlType, size, isNullable, false));
		}
		sqlColumns.close();
		
		// Load primary key information
		ResultSet sqlPK = metaData.getPrimaryKeys(catalog, schema, table.getName());
		Set<String> pkColumns = new LinkedHashSet<String>();
		String pkName = null;
		while (sqlPK.next())
		{
			if (pkName == null) pkName = makeSafe(sqlPK.getString("PK_NAME"));
			pkColumns.add(makeSafe(sqlPK.getString("COLUMN_NAME")));
		}
		if (pkName != null || pkColumns.size() > 0)
		{
			pkInfo = new PrimaryKeyInfo(pkName, pkColumns);
		}
		sqlPK.close();
		
		// Load foreign keys
		ResultSet sqlFK = metaData.getImportedKeys(catalog, schema, table.getName());
		Map<String, ForeignKeyInfo> fkInfo = new HashMap<String, ForeignKeyInfo>();
		while (sqlFK.next())
		{
			String fkName = makeSafe(sqlFK.getString("FK_NAME"));
			String fkColumn = makeSafe(sqlFK.getString("FKCOLUMN_NAME"));
			String ref = makeSafe(sqlFK.getString("PKTABLE_NAME"));
			String refColumn = makeSafe(sqlFK.getString("PKCOLUMN_NAME"));
			
			ForeignKeyInfo fk = fkInfo.get(fkName);
			if (fk == null)
			{
				fk = new ForeignKeyInfo(fkName, ref);
				fkInfo.put(fkName, fk);
				foreignKeys.add(fk);
			}
			fk.fkColumns.add(fkColumn);
			fk.refColumns.add(refColumn);
		}
		sqlFK.close();
		
		// Load indexes
		ResultSet sqlIndexes = metaData.getIndexInfo(catalog, schema, table.getName(), false, true);
		Map<String, IndexInfo> indexInfo = new HashMap<String, IndexInfo>();
		while (sqlIndexes.next())
		{
			String indexName = makeSafe(sqlIndexes.getString("INDEX_NAME"));
			String column = makeSafe(sqlIndexes.getString("COLUMN_NAME"));
			boolean isUnique = !sqlIndexes.getBoolean("NON_UNIQUE");
			IndexInfo ii = indexInfo.get(indexName);
			if (ii == null)
			{
				ii = new IndexInfo(indexName, isUnique);
				indexInfo.put(indexName, ii);
				indexes.add(ii);
			}
			ii.columns.add(column);
		}
		sqlIndexes.close();
	}
	
	/**
		Create table info by reading the Hibernate mappings.
		@param table The Hibernate table object
		@param dialect The current database dialect
	*/
	@SuppressWarnings("unchecked")
	public TableInfo(Table table, Dialect dialect)
	{
		this(table);
		this.dialect = dialect;
		
		// Load column information
		Iterator<Column> hibColumns = (Iterator<Column>)table.getColumnIterator();
		while (hibColumns.hasNext())
		{
			Column column = hibColumns.next();
			String name = column.getName();
			Integer sqlType = column.getSqlTypeCode();
			String type = "";
			try
			{
				type = column.getSqlType(dialect, null);
			}
			catch (Exception ex)
			{}
			int size = column.getLength();
			boolean isUnique = column.isUnique();
			boolean isNullable = column.isNullable();
			columns.add(new ColumnInfo(name, type, sqlType, size, isNullable, isUnique));
			if (isUnique)
			{
				// Single columns that are UNIQUE are NOT included in Table.getUniqueKeyIterator()
				// so we need to define it here
				IndexInfo uk = new IndexInfo(name, true);
				uk.columns.add(name);
				indexes.add(uk);
			}
		}
		
		//	Load primary key information
		PrimaryKey pk = table.getPrimaryKey();
		Set<String> pkColumns = new LinkedHashSet<String>();
		String pkName = null;
		if (pk != null)
		{
			pkName = pk.getName();
			hibColumns = pk.getColumnIterator();
			while (hibColumns.hasNext())
			{
				Column column = hibColumns.next();
				pkColumns.add(column.getName());
			}
		}
		if (pkName != null || pkColumns.size() > 0)
		{
			pkInfo = new PrimaryKeyInfo(pkName, pkColumns);
		}

		// Load indexes
		Iterator<Index> indexIterator = table.getIndexIterator();
		while (indexIterator.hasNext())
		{
			Index index = indexIterator.next();
			String name = index.getName();

			IndexInfo ii = new IndexInfo(name, false);
			hibColumns = index.getColumnIterator();
			while (hibColumns.hasNext())
			{
				Column column = hibColumns.next();
				ii.columns.add(column.getName());
			}
			if (findIndexName(null, ii.columns) == null)
			{
				// Do not add duplicates
				indexes.add(ii);
			}
		}
		
		// Load unique constraints (multi-column only!)
		Iterator<UniqueKey> ukIterator = table.getUniqueKeyIterator();
		while (ukIterator.hasNext())
		{
			UniqueKey uk = ukIterator.next();
			String name = uk.getName();

			IndexInfo ii = new IndexInfo(name, true);
			hibColumns = uk.getColumnIterator();
			while (hibColumns.hasNext())
			{
				Column column = hibColumns.next();
				ii.columns.add(column.getName());
			}
			if (findIndexName(null, ii.columns) == null)
			{
				// Do not add duplicates
				indexes.add(ii);
			}
		}
		
		// Load foreign keys
		Iterator<ForeignKey> fkIterator = table.getForeignKeyIterator();
		while (fkIterator.hasNext())
		{
			ForeignKey fk = fkIterator.next();
			String fkName = fk.getName();
			String refName = fk.getReferencedTable().getName();

			ForeignKeyInfo fki = new ForeignKeyInfo(fkName, refName);
			hibColumns = fk.getColumnIterator();
			while (hibColumns.hasNext())
			{
				Column column = hibColumns.next();
				fki.fkColumns.add(column.getName());
			}
			
			if (findIndexName(null, fki.fkColumns) == null && !pkColumns.equals(fki.fkColumns))
			{
				// Define an implicit index for the foreign key columns
				// if there is no explicit index (or primary key) for the same set of columns
				// If it is a single column, use the name of that column
				// otherwise the name from the foreign key
				IndexInfo fkIndex = new IndexInfo(fki.fkColumns.size() == 1 ? fki.fkColumns.iterator().next() : fkName, false);
				fkIndex.columns.addAll(fki.fkColumns);
				indexes.add(fkIndex);
			}
			
			List<Column> refColumns = fk.isReferenceToPrimaryKey() ?
					fk.getReferencedTable().getPrimaryKey().getColumns() :
					fk.getReferencedColumns();
					
			for (Column column : refColumns)
			{
				fki.refColumns.add(column.getName());
			}
			
			foreignKeys.add(fki);
		}
		

	}
	
	/**
		Get the table this information is representing.
		@return The table object
		@since 3.1
	*/
	public Table getTable()
	{
		return table;
	}
	
	/**
		Check if the table exists in the database or not. This value is only
		meaningful if the TableInfo object was created from JDBC metadata
		using {@link #TableInfo(Table, DatabaseMetaData)}.
		@since 3.1
	*/
	public boolean existsInDb()
	{
		return existsInDb;
	}
	
	/**
		Sometime the database put quotes around names. This method strips all
		extra characters.
		@param name The name to make safe, or null
		@return The safe name or null
		@since 2.1.1
	*/
	private String makeSafe(String name)
	{
		if (name == null) return null;
		return name.replaceAll("\\W", "");
	}
	
	/**
		Get information about the columns in the table.
		@return A set containing information about each column
	*/
	public Set<ColumnInfo> getColumns()
	{
		return Collections.unmodifiableSet(columns);
	}
	
	/**
		Get information about the primary key of the table
		@return A PrimaryKeyInfo object, or null if the table doesn't have a primary key
	*/
	public PrimaryKeyInfo getPrimaryKey()
	{
		return pkInfo;
	}
	
	/**
		Get information about the foreign keys in the table.
		@return A set containing information about each foreign key
	*/
	public Set<ForeignKeyInfo> getForeignKeys()
	{
		return Collections.unmodifiableSet(foreignKeys);
	}
	
	/**
		Check if there is a foreign key with the same name or exactly the same 
		set of columns referencing the same table.
	
		@param name The name of the foreign key to look for, or null to only use the
			columns
		@param columns A set containing the column names that should be present in 
			the foreign key
		@param refName The referenced table
		@return The name of the foreign key, or null if no foreign key was found
		@since 3.8
	*/
	public String findForeignKeyName(String name, Set<String> columns, String refName)
	{
		for (ForeignKeyInfo fkInfo : foreignKeys)
		{
			if (name != null && name.equals(fkInfo.getName())) return name;
			if (columns.equals(fkInfo.getFkColumns()) && refName.equalsIgnoreCase(fkInfo.getRefName()))
			{
				return fkInfo.getName();
			}
		}
		return null;
	}

	/**
		Get information about the indexes in the table. Note! Some databases
		ie. MySQL, uses indexes to keep track of foreign and primary keys. This means 
		that the information in this set includes one index for each primary/foreign key
		in addition to those indexes created for other reasons.
		
		@return A set containing information about each index
	*/
	public Set<IndexInfo> getIndexes()
	{
		return Collections.unmodifiableSet(indexes);
	}
	
	/**
		It is safe to drop an index if it is not the primary key.
		
		@param ii The IndexInfo object we want to check if it is safe to drop
		@return TRUE if is safe to drop the index, FALSE otherwise
	 */
	public boolean safeToDrop(IndexInfo ii)
	{
		// Check if the name matches...
		String name = ii.getName();
		
		// ... the primary key
		if (name.equals(pkInfo == null ? null : pkInfo.getName())) return false;
		
		return true;
	}
	
	/**
		Check if there is an index with the same name or exactly the same 
		set of columns.
		<p>
		Hibernate doesn't generate a name for multi-column indexes or unique 
		constraints, which means that we have to find any index matching
		the column names. In case, Hibernate starts generating the indexes
		with names, we check against the name as well.

		@param name The name of the index to look for, or null to only use the
			columns
		@param columns A set containing the column names that should be present in 
			the index
		@return The name of the index, or null if no index was found
	*/
	public String findIndexName(String name, Set<String> columns)
	{
		for (IndexInfo iiInfo : indexes)
		{
			if (name != null && name.equals(iiInfo.getName())) return name;
			if (columns.equals(iiInfo.getColumns())) return iiInfo.getName();
		}
		return null;
	}
	
	/**
		Find all indexes with the given set of columns. Used for
		getting rid of duplicates.
		@param columns A set containing the column names that should be present in 
			the index
		@return A list with the index information or an empty list of no index is found
		@since 3.4
	*/
	public List<IndexInfo> findAll(Set<String> columns)
	{
		List<IndexInfo> all = new ArrayList<IndexInfo>();
		for (IndexInfo iiInfo : indexes)
		{
			if (columns.equals(iiInfo.getColumns())) all.add(iiInfo);
		}
		return all;
	}

	/**
		Find information about a column with the given name.
		@param column The name of the column
		@return The column information or null if no column is found
		@since 3.1
	*/
	public ColumnInfo findColumn(String column)
	{
		for (ColumnInfo ciInfo : columns)
		{
			if (ciInfo.getName().equals(column)) 
			{
				return ciInfo;
			}
		}
		return null;
	}
	
	/**
		Get the fully quialified and quoted table name using
		the quoting rules from the given DBEngine
		@param engine A DbEngine
		@return The fully quoted table name
		@since 3.1
	*/
	public String getFullQuotedTableName(DbEngine engine)
	{
		StringBuilder name = new StringBuilder();
		if (table.getCatalog() != null) 
		{
			name.append(engine.getQuotedName(table.getCatalog())).append(".");
		}
		if (table.getSchema() != null) 
		{
			name.append(engine.getQuotedName(table.getSchema())).append(".");
		}
		name.append(engine.getQuotedName(table.getName()));
		return name.toString();
	}
	
	/**
		Count the number of rows in the table.
		@param connection The connection to user for counting
		@since 3.1
	*/
	public long getRowCount(DbEngine engine, Connection connection, boolean approximate)
		throws SQLException
	{
		if ((numberOfRows < 0 && !approximate) || (approximateNumberOfRows < 0 && approximate))
		{
			
			Statement st = null;
			ResultSet result = null;
			try
			{
				st = connection.createStatement();
				String sql = null;
				if (approximate)
				{
					sql = engine.getApproximateRowCountSql(table.getCatalog(), table.getSchema(), table.getName());
				}
				if (sql == null)
				{
					sql = "SELECT count(*) FROM " + getFullQuotedTableName(engine);
				}
				result = st.executeQuery(sql);
				result.next();
				long count = result.getLong(1);
				if (approximate)
				{
					approximateNumberOfRows = count;
				}
				else 
				{
					numberOfRows = count;
				}
			}
			finally
			{
				if (result != null) result.close();
				if (st != null) st.close();
			}
		}
		return approximate ? approximateNumberOfRows : numberOfRows;
	}
	
	private static final Set<Integer> STRING_TYPE = new HashSet<Integer>(
			Arrays.asList(
				Types.CHAR, Types.CLOB, Types.LONGNVARCHAR, Types.LONGVARCHAR, 
				Types.NCHAR, Types.NCLOB, Types.NVARCHAR, Types.VARCHAR
			));
	
	/**
	 	Holds information about a single column in a table.
	*/
	public static class ColumnInfo
		implements Comparable<ColumnInfo>
	{
		private final String name;
		private final String type;
		private final Integer sqlType;
		private final int size;
		private final boolean isNullable;
		private final boolean isUnique;
		
		/**
			Create a new ColumnInfo object.
			
			@param name The name of the column
			@param type The type of the column (int, float, varchar, ...) etc.
			@param size The size of the column
			@param isNullable If the column is nullable
			@param isUnique If the column must have unique values
		*/
		public ColumnInfo(String name, String type, Integer sqlType, int size, boolean isNullable, boolean isUnique)
		{
			this.name = name;
			this.type = type;
			this.sqlType = sqlType;
			this.size = size;
			this.isNullable = isNullable;
			this.isUnique = isUnique;
		}
		
		public String getName()
		{
			return name;
		}
		public String getType()
		{
			return type;
		}
		/**
			@since 2.9
		*/
		public Integer getSqlType()
		{
			return sqlType;
		}
		public int getSize()
		{
			return size;
		}
		public boolean isNullable()
		{
			return isNullable;
		}
		public boolean isUnique()
		{
			return isUnique;
		}
		
		/**
			@since 2.9
		*/
		public boolean isStringType()
		{
			return STRING_TYPE.contains(sqlType);
		}
		
		@Override
		public String toString()
		{
			return name + " " + type + "(" + size + ")" + (isNullable ? " NULL" : " NOT NULL") + 
				(isUnique ? " UNIQUE" : "");
		}

		@Override
		public int compareTo(ColumnInfo o)
		{
			return name.compareTo(o.name);
		}
	}
	
	/**
	 	Holds information about the primary key of a table.
	*/
	public static class PrimaryKeyInfo
	{
		private final String name;
		private final Set<String> columns;
		
		/**
			Create a new PrimaryKeyInfo object.
			
			@param name The name of the primary
			@param columns The columns that are part of the primary key
		 */
		public PrimaryKeyInfo(String name, Set<String> columns)
		{
			this.name = name;
			this.columns = columns;
		}
		
		public String getName()
		{
			return name;
		}
		public Set<String> getColumns()
		{
			return columns;
		}
		
		@Override
		public String toString()
		{
			return columns.isEmpty() ? "MISSING" : (name == null ? "<no name>" : name) + " " + columns;
		}
	}
	
	/**
	 	Holds information about a foreign key of a table.
	*/
	public static class ForeignKeyInfo
	{
		private final String name;
		private final String refName;
		private final Set<String> fkColumns;
		private final Set<String> refColumns;
		
		/**
			Create a new ForeignKeyInfo object. You should also
			add column names to the {@link #getFkColumns()} and {@link #getRefColumns()}
			sets.
			
			@param name The name of the foreign key
			@param refName The referenced table
		*/
		public ForeignKeyInfo(String name, String refName)
		{
			this.name = name;
			this.refName = refName;
			this.fkColumns = new LinkedHashSet<String>();
			this.refColumns = new LinkedHashSet<String>();
		}
		
		public String getName()
		{
			return name;
		}
		public String getRefName()
		{
			return refName;
		}
		public Set<String> getFkColumns()
		{
			return fkColumns;
		}
		public Set<String> getRefColumns()
		{
			return refColumns;
		}
		
		@Override
		public String toString()
		{
			return name + " " + fkColumns + " REFERENCES " + refName + " " + refColumns;
		}
	}
	
	/**
	 	Holds information about an index in a table.
	*/
	public static class IndexInfo
	{
		private final String name;
		private final boolean isUnique;
		private final Set<String> columns;
		
		/**
			Create a new IndexInfo object. You should also add column names
			to the {@link #getColumns()} set.
			
			@param name The name of the index
			@param isUnique If the index must have unique values or not
		 */
		public IndexInfo(String name, boolean isUnique)
		{
			this.name = name;
			this.isUnique = isUnique;
			this.columns = new LinkedHashSet<String>();
		}
		
		public String getName()
		{
			return name;
		}
		public boolean isUnique()
		{
			return isUnique;
		}
		public Set<String> getColumns()
		{
			return columns;
		}
		
		@Override
		public String toString()
		{
			return name + " " + columns;
		}
	
	}
}
