/**
	$Id: MetadataParser.java 5319 2010-04-20 08:02:16Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.bfs;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Pattern;

import net.sf.basedb.core.File;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.encode.EncoderDecoder;
import net.sf.basedb.util.encode.TabCrLfEncoderDecoder;
import net.sf.basedb.util.parser.FlatFileParser;

/**
	Parser implementation that parses a BFS metadata file. Before parsing
	is started the stream to parse must be specified by calling
	{@link #setInputStream(InputStream)}. The actual parsing can be done
	in two different ways:
	
	<ul>
	<li>"Manual" parsing, using the {@link #parseToBof()}, {@link #nextSection()}
		and {@link #nextEntry()} methods.
	<li>Event-based parsing using the {@link #parse(EventHandler)} method.
		This parser issues the following events:
		<ul>
		<li>{@link #BOF_EVENT}: When the beginning-of-file marker is found.
		<li>{@link #SECTION_EVENT}: When the start of a new section is found.
		<li>{@link #ENTRY_EVENT}: For each entry in a section.
		<li>{@link #SECTION_END_EVENT}: After the last entry in a section.
		<li>{@link #END_OF_FILE_EVENT}: When the end of the file has been reached.
		</ul>
		The {@link MetadataModel} class implements a simple event handler that 
		collects the information from the metadata file and provides methods
		for accessing it.
	</ul>
		
	<p>
	This class may be subclassed to provide customized behaviour.
	
	@author Nicklas
	@version 2.15
	@base.modified $Date: 2010-04-20 10:02:16 +0200 (ti, 20 apr 2010) $
*/
public class MetadataParser
	implements BfsParser
{

	/**
		Event type that is issued when the beginning-of-file marker is found.
		The event data is a string with the BFS subtype, or null if no subtype
		was defined.
	*/
	public static final EventType<String> BOF_EVENT = 
		new EventType<String>("metadata.bof");
	
	/**
		Event type that is issued for each new section that is found in the file.
		The event data is a string with the name (decoded) of the section.
	*/
	public static final EventType<String> SECTION_EVENT = 
		new EventType<String>("metadata.section");
	
	/**
		Event type that is issued for each entry in a section in the order they
		are present in the file. The event data is a {@link MetadataModel.SectionEntry}
		object with the decoded key and values.
	*/
	public static final EventType<MetadataModel.SectionEntry> ENTRY_EVENT = 
		new EventType<MetadataModel.SectionEntry>("metadata.entry");
	
	/**
		Event type that is issued at the end of each section in the file.
		No event data is submitted.
	*/
	public static final EventType<String> SECTION_END_EVENT = 
		new EventType<String>("metadata.section.end");
	
	/**
		Event type that is issued when the end-of-file had been reached.
		No event data is submitted.
	*/
	public static final EventType<Object> END_OF_FILE_EVENT = 
		new EventType<Object>("metadata.end");

	
	/**
		Utility method for creating a metadata parser when you have an
		input stream.
		@param in The input stream the parser should read from
		@param filename Optional, the name of the file the input stream
			is reading from
	*/
	public static MetadataParser create(InputStream in, String filename, long fileSize)
	{
		MetadataParser parser = new MetadataParser();
		parser.setFilename(filename);
		parser.setInputStream(in);
		parser.setFileSize(fileSize);
		return parser;
	}
	
	/**
		Utility method for creating a metadata parser for a file in the BASE
		file system.
		@param file The file in the BASE file system
	*/
	public static MetadataParser create(File file)
	{
		return create(file.getDownloadStream(0), file.getName(), file.getSize());
	}
	
	/**
		Utility method for creating a metadata parser for a file in the local
		file system.
		@param file The file in the local file system
	*/
	public static MetadataParser create(java.io.File file)
		throws IOException
	{
		return create(new FileInputStream(file), file.getName(), file.length());
	}
	
	private final EncoderDecoder decoder;
	private InputStream in;
	private String filename;
	private long fileSize;
	private FlatFileParser ffp;
	private List<MetadataModel.SectionEntry> sectionEntries;
	
	/**
		Create a new metadata parser.
	*/
	public MetadataParser()
	{
		this.decoder = new TabCrLfEncoderDecoder(true);
	}
	
	/**
		Decode an encoded value. Metadata values are encoded/decoded
		with a {@link TabCrLfEncoderDecoder}.
		
		@param value The encoded value
		@return The decoded value
	*/
	public String decodeValue(String value)
	{
		return decoder.decode(value);
	}

	/**
		Get the file name that this writer is printing to.
		@return The file name or null if not known
	*/
	@Override
	public String getFilename()
	{
		return filename;
	}
	
	/**
		Set the file name that this writer is printing to.
	*/
	public void setFilename(String filename)
	{
		this.filename = filename;
	}
	
	/**
		Get the size in bytes of the file that this parser is reading from.
		@return The size or -1 if not known
	*/
	@Override
	public long getFileSize()
	{
		return fileSize;
	}
	
	/**
		Set the size of the file or -1 if not known.
	*/
	public void setFileSize(long fileSize)
	{
		this.fileSize = fileSize;
	}

	
	@Override
	public void close() 
	{
		if (in != null) FileUtil.close(in);
	}

	/**
		Set the input stream that should be parsed. This will also reset the
		parser. All information from a previously parsed file is lost.
		@param in The stream to parse (in UTF-8 format)
	*/
	public void setInputStream(InputStream in)
	{
		close();
		this.in = in;
		this.ffp = null;
		this.sectionEntries = null;
	}
	
	/**
		Parse the input stream and notify the specified event handler with events.
		The event handler is responsible for keeping track of and storing the
		data of interest.
		
		@param handler An event handler
		@throws IOException If there is an error reading the stream data
		@throws NullPointerException If the stream or event handler is null
	*/
	public void parse(EventHandler handler)
		throws IOException
	{
		if (handler == null) throw new NullPointerException("handler");
		
		String subType = parseToBof();
		handleBof(handler, subType);
		
		// Repeat as long as there are more sections in the file
		String section = nextSection();
		while (section != null)
		{
			ThreadSignalHandler.checkInterrupted();
			handleSection(handler, section);
			MetadataModel.SectionEntry entry = nextEntry();
			while (entry != null)
			{
				handleEntry(handler, entry);
				entry = nextEntry();
			}
			handleSectionEnd(handler, section);
			section = nextSection();
		}
		handleEndOfFile(handler);
	}
	
	/**
		Parse the start of the file until the beginning-of-file marker
		(BFSfile) is found. This method can only be called once on 
		a given input stream. To parse a second file, call {@link 
		#setInputStream(InputStream)} again to reset the parser.
		
		@return The subtype of the BFS or null if no subtype is found
		@throws IOException If the beginning-of-file marker is not found
			 or if there is any other IO error
		@throws IllegalStateException If the parsing has already started
		@throws NullPointerException If no input stream has been specified
	*/
	public String parseToBof()
		throws IOException
	{
		if (in == null) throw new NullPointerException("inputStream");
		if (ffp != null) 
		{
			throw new IllegalStateException("Parsing of file '" + getFilename() + 
				"' has already started. Call setInputStream() to reset the parser.");
		}
		ffp = createFlatFileParser(in);
		// Parse to the beginning-of-file marker
		if (!ffp.parseToBof())
		{
			throw new IOException("Can't find 'BFSformat' in file");
		}
		return decodeValue(ffp.getBofType());
	}
	
	/**
		Parse to the next section in the metadata file. This method can be called
		repeatedly until null is returned which indicates that the end of file has
		been reached.
		
		@return The name of the section, or null if no more sections could be found
		@throws IOException If there is an IO error during the parsing
		@throws IllegalStateException If the parsing has not been started. 
			{@link #parseToBof()} must be called before this method is called.
	*/
	public String nextSection()
		throws IOException
	{
		if (ffp == null) 
		{
			throw new IllegalStateException("Parsing of file '" + getFilename() + 
				"' hasn't started. Call parseToBof() to begin parsing.");
		}
		if (!ffp.hasMoreSections()) return null;
		
		FlatFileParser.Line sectionLine = ffp.nextSection();
		String section = decodeValue(sectionLine.name());
		ffp.parseHeaders();
		sectionEntries = new LinkedList<MetadataModel.SectionEntry>();
		for (FlatFileParser.Line l : ffp.getLines())
		{
			if (l.type() == FlatFileParser.LineType.HEADER)
			{
				String key = decodeValue(l.name());
				String[] values = l.value().split("\\t", -1);
				for (int i = 0; i < values.length; ++i)
				{
					values[i] = decodeValue(values[i]);
				}
				sectionEntries.add(new MetadataModel.SectionEntry(key, values));
			}
		}
		return section;
	}
	
	/**
		Get the next entry in the current section. This method can be called
		repeatedly until null is returned which indicates that the end of the
		section has been reached.
		
		@return The key and values or the entry, or null if no more entries could be found
		@throws IOException If there is an IO error during the parsing
		@throws IllegalStateException If the parsing has not been started. 
			{@link #parseToBof()} must be called before this method is called.
	*/
	public MetadataModel.SectionEntry nextEntry()
		throws IOException
	{
		if (ffp == null) 
		{
			throw new IllegalStateException("Parsing of file '" + getFilename() + 
				"' hasn't started. Call parseToBof() to begin parsing.");
		}
		if (sectionEntries == null || sectionEntries.size() == 0) return null;
		MetadataModel.SectionEntry entry = sectionEntries.remove(0);
		return entry;
	}
	
	@Override
	public int getCurrentLine()
	{
		return ffp == null ? -1 : ffp.getParsedLines();
	}
	
	@Override
	public long getParsedBytes()
	{
		return ffp == null ? -1 : ffp.getParsedBytes();
	}
	
	/**
		Create a new flat file parser that can parse BFS metadata
		files. This method should set all regular expressions that
		are needed to parse the stream (assumed to be UTF-8).
	*/
	protected FlatFileParser createFlatFileParser(InputStream in)
	{
		FlatFileParser ffp = new FlatFileParser();
		ffp.setBofMarkerRegexp(Pattern.compile("BFSformat\t?(.*)"));
		ffp.setIgnoreRegexp(Pattern.compile("#.*|\\s*"));
		ffp.setSectionRegexp(Pattern.compile("\\[(.*)\\]\\s*"));
		ffp.setHeaderRegexp(Pattern.compile("(.+?)\\t(.*)"));
		ffp.setInputStream(in, "UTF-8");
		return ffp;
	}
	
	/**
		Handle the beginning-of-file marker event. The default
		implementation sends a {@link #BOF_EVENT} notification to the 
		event handler.

		@param handler The event handler from the {@link #parse(EventHandler)} method
		@param subtype The decoded subtype or null if no subtype was specified in the file
	*/
	protected void handleBof(EventHandler handler, String subtype)
	{
		handler.handleEvent(BOF_EVENT, subtype, this);
	}
	
	/**
		Handle the new section event. The default implementation sends a 
		{@link #SECTION_EVENT} notification to the event handler.
	
		@param handler The event handler from the {@link #parse(EventHandler)} method
		@param name The decoded name of the section
	*/
	protected void handleSection(EventHandler handler, String name)
	{
		handler.handleEvent(SECTION_EVENT, name, this);
	}
	
	/**
		Handle the section entry event. The default implementation sends a 
		{@link #ENTRY_EVENT} notification to the event handler.
	
		@param handler The event handler from the {@link #parse(EventHandler)} method
		@param entry Key and values from the section entry
	*/
	protected void handleEntry(EventHandler handler, MetadataModel.SectionEntry entry)
	{
		handler.handleEvent(ENTRY_EVENT, entry, this);
	}
	
	/**
		Handle the end-of-section event. The default implementation sends a 
		{@link #SECTION_END_EVENT} notification to the event handler.
	
		@param handler The event handler from the {@link #parse(EventHandler)} method
		@param name The decoded name of the section
	*/
	protected void handleSectionEnd(EventHandler handler, String name)
	{
		handler.handleEvent(SECTION_END_EVENT, name, this);
	}

	
	/**
		Handle the end-of-file event. The default implementation sends an
		{@link #END_OF_FILE_EVENT} notification to the event handler.
		@param handler The event handler from the {@link #parse(EventHandler)} method
	*/
	protected void handleEndOfFile(EventHandler handler)
	{
		handler.handleEvent(END_OF_FILE_EVENT, null, this);
	}
}
