/**
	$Id: FieldConverter.java 5277 2010-03-22 12:45:37Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.export.spotdata;


import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ExtraValue.CoordinateType;

/**
	Exportable field converter that support BASE 1 naming conventions.
	This implementation rely in most cases on formulas with the same 
	name as the BASE 1 field. BASE 1 fields that happens to have the 
	same name in BASE 2 doesn't need formulas.
	<p>
	This class extends the standard field converter and only overrides
	methods that is used by the {@link net.sf.basedb.plugins.Base1PluginExecuter} 
	so thatit can work as expected. The major differences are the name of
	intensity fields and the _xc_ prefix on extra value fields.
	
	@author Nicklas
	@version 2.15
	@base.modified $Date: 2010-03-22 13:45:37 +0100 (må, 22 mar 2010) $
*/
public class Base1FieldConverter
	extends StandardFieldConverter
{
	
	/**
		Create a new converter for the given bioassay set.
		@param dc A DbControl to use for database access
		@param source The source bioassay set
	*/
	public Base1FieldConverter(DbControl dc, BioAssaySet source)
	{
		super(dc, source);
	}
	
	/*
		From the ExportableFieldConverter interface
		-------------------------------------------
	*/

	/**
		Convert a reporter field. The conversion is done in steps
		until a match is found:
		<ul>
		<li>If the field starts with '_xc_' it could be an extra value. We
			check if there is a position extra value with an external id that is the
			same as the rest of the field (eg. without the _xc_ prefix).
		<li>Check if a formula that has the exactly the same name as the
			given field name exists. The formula must be a column expression
			and it must usable on the raw data type and the intensity transform 
			that the source bioassay set uses.
		<li>Check if an extended reporter property with exactly the same name
			as the given field name exists.
		</ul>
		@param name The BASE 1 field name
		@param required Controls if null is returned or an exception is thrown when
			a field is not found
		@return A DynamicField expression or null if the field name can't be 
			converted
	*/
	@Override
	public DynamicField getReporterField(String name, boolean required) 
	{
		DynamicField fld = null;
		if (name.startsWith("_xc_"))
		{
			fld = getExtraValueField(name.substring(4), CoordinateType.POSITION, name);
		}
		if (fld == null)
		{
			fld = getFormulaField(name, name);
		}
		if (fld == null)
		{
			fld = getReporterProperty(name, name);
		}
		if (fld == null && required)
		{
			throw new ItemNotFoundException("Unknown reporter field: " + name +
				" (you can add this field as an formula)");
		}
		return fld;
	}

	/**
		Convert a spot field. The conversion is done in steps
		until a match is found:
		<ul>
		<li>If the field starts with '_xc_' it could be an extra value. We
			check if there is an spot extra value with an external id that is the
			same as the rest of the field (eg. without the _xc_ prefix).
		<li>Check if a formula that has the exactly the same name as the
			given field name exists. The formula must be a column expression
			and it must usable on the raw data type and the intensity transform 
			that the source bioassay set uses.
		<li>Check if a raw data property with exactly the same name
			as the given field name exists.
		<li>Check if the field name is 'intensityX' where 'X' is a channel
			number.
		</ul>
		@param name The BASE 1 field name
		@param required Controls if null is returned or an exception is thrown when
			a field is not found
		@return A DynamicField expression or null if the field name can't be 
			converted
	*/
	@Override
	public DynamicField getSpotField(String name, boolean required) 
	{
		DynamicField fld = null;
		if (name.startsWith("_xc_"))
		{
			fld = getExtraValueField(name.substring(4), CoordinateType.SPOT, name);
		}
		if (fld == null)
		{
			fld = getFormulaField(name, name);
		}
		if (fld == null)
		{
			fld = getRawDataProperty(name, name);
		}
		if (fld == null)
		{
			fld = getIntensityField(name, name);
		}
		if (fld == null && required)
		{
			throw new ItemNotFoundException("Unknown spot field: " + name +
				" (you can add this field as a formula)");
		}		
		return fld;
	}

	// --------------------------------

	/**
		Create a dynamic field for spot intensity if the field name has
		the pattern 'intensityX' where 'X' is a channel number between 1
		and the number of channels in the raw data type used by the 
		source bioassay set.
		@param fieldName The field name
		@return A dynamic field, or null if the name doesn't match the pattern
	*/
	@Override
	protected DynamicField getIntensityField(String fieldName, String exportTitle)
	{
		DynamicField df = null;
		for (int i = 1; i <= getSource().getRawDataType().getChannels(); ++i)
		{
			if (fieldName.equals("intensity"+i))
			{
				df = ExportableFieldFactory.channel(i, exportTitle, null);
				break;
			}
		}
		return df;
	}

}
