/**
	$Id: JobAgentConnection.java 6898 2015-05-12 09:27:08Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.jobagent;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sf.basedb.util.SocketUtil;

/**
	This class is used by client applications to communicate with job agents. 
	It used to send commands to job agents. For example, it can send
	<code>start</code>, <code>stop</code> and <code>info</code> commands.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2015-05-12 11:27:08 +0200 (ti, 12 maj 2015) $
	@see JobAgentServerConnection
*/
public class JobAgentConnection
{

	/**
		Regexp used to parse a typical answer.
	*/
	private static final Pattern HEADER_REGEXP = Pattern.compile("(.*):(.*)");
	
	/**
		Utility method for parsing a 'typical' answer from a job agent.
		The typical answer is a string that contains a key-value pair on each
		line, separated by a colon. This method allows multiple entries for
		the same key.
		
		@param answer The answer
		@return A map with the key as index and a list with all values
		@since 2.16
	*/
	public static Map<String, List<String>> parseAnswer(String answer)
	{
		Map<String, List<String>> headers = new HashMap<String, List<String>>();
		Matcher m = HEADER_REGEXP.matcher(answer);
		while (m.find())
		{
			String header = m.group(1);
			String value = m.group(2);
			List<String> values = headers.get(header);
			if (values == null)
			{
				values = new ArrayList<String>();
				headers.put(header, values);
			}
			values.add(value);
		}
		return headers;
	}

	/**
		Utility method for parsing a 'typical' answer from a job agent.
		The typical answer is a string that contains a key-value pair on each
		line, separated by a colon. This method allows only a single entry for
		the same key.
		
		@param answer The answer
		@return A map with the key as index to the value
		@since 2.16
	*/
	public static Map<String, String> parseSimpleAnswer(String answer)
	{
		Map<String, String> headers = new HashMap<String, String>();
		Matcher m = HEADER_REGEXP.matcher(answer);
		while (m.find())
		{
			String header = m.group(1);
			String value = m.group(2);
			headers.put(header, value);
		}
		return headers;
	}
	
	
	private final String server;
	private final int port;
	private final int timeout;
	
	/**
		Create a connection to the local job agent running on the specified port.
		@param port The port the job agent is listening on
		@param timeout The timeout in milliseconds
	*/
	public JobAgentConnection(int port, int timeout)
	{
		this(null, port, timeout);
	}
	
	/**
		Create a connection to the job agent running on the specified server
		and listening on the specified port.
		@param server The server the job agent is running on, or null
			to connect to the local host
		@param port The port the job agent is listening on
		@param timeout The timeout in milliseconds
	*/
	public JobAgentConnection(String server, int port, int timeout)
	{
		this.server = server;
		this.port = port;
		this.timeout = timeout;
	}
	
	/**
		Send the <code>ping</code> command to the job agent. The only
		answer to this command is <code>OK</code>.
		@return OK if successful
		@throws IOException If there is an error
	 */
	public String sendPing()
		throws IOException
	{
		return send("ping");
	}

	/**
		Send a <code>start</code> request to the job agent.
		@return OK if successful
		@throws IOException If there is an error
	*/
	public String sendStart()
		throws IOException
	{
		return send("start");
	}
	
	public String sendStop()
		throws IOException
	{
		return send("stop");
	}
	
	public String sendPause()
		throws IOException
	{
		return send("pause");
	}
	
	/**
		Send an <code>info</code> or <code>status</code> request to the job agent.
		The difference is that a <code>status</code> request doesn't return
		information about cpu, memory or executing jobs.
		
		@param full If TRUE a <code>info</code> request is sent, othwerwise
			a <code>status</code> request is sent
		@return A <code>JobAgentInfo</code> object containing the 
			information
		@throws IOException If there is an error
	*/
	public JobAgentInfo getInfo(boolean full)
		throws IOException
	{
		return new JobAgentInfo(send(full ? "info" : "status"));
	}
	
	/**
		Close the connection to the job agent. In the current implementation
		all connections are stateless and this method does nothing. This might change
		in the future if a more advanced protocol is implemented. Clients should
		always call this method.
	*/
	public void close()
	{}
	
	/**
		Send a remote control command to the job agent.
		
		@param cmd The command to send
		@return The answer
		@throws IOException If there is an error
		@since 2.16 (was private before that)
	*/
	public String send(String cmd)
		throws IOException
	{
		Socket socket = null;
		String answer;
		try
		{
			socket = connect();
			answer = sendCmd(socket, cmd);
		}
		finally
		{
			SocketUtil.close(socket);
		}
		return answer;
	}
	
	/**
		Create a socket connection to the job agent.
		@return The socket
	*/
	private Socket connect()
		throws IOException
	{
		Socket s = new Socket();
		InetSocketAddress address = null;
		if (server == null)
		{
			address = new InetSocketAddress(InetAddress.getLocalHost(), port);
		}
		else
		{
			address = new InetSocketAddress(server, port);
		}
		s.connect(address, timeout);
		return s;
	}

	/**
		Send a command to the job agent and return the answer. If the answer is
		<code>FAILED</code> an exception will be thrown.
		
		@param socket The socket used for communicating with the job agent
		@param cmd The command to send
		@return The answer from the job server
		@throws IOException If there is a communications error
	*/
	private String sendCmd(Socket socket, String cmd)
		throws IOException
	{
		// Send cmd to the job agent
		SocketUtil.send(socket, cmd, true);
		
		// Read the answer
		String answer = SocketUtil.read(socket, true);
		
		// The answer must start with OK, or it is an error
		if (answer == null || answer.length() == 0)
		{
			throw new IOException("No answer from "+socket.getInetAddress().toString());
		}
		else if (answer.indexOf("OK") != 0)
		{
			throw new IOException(answer.toString());
		}
		return answer.toString();
	}

}
